""" Step that builds a docker image from its specification in Dockerfile-mpl. """

from logging import Logger

from .docker_after_build import AfterBuildDocker
from .. import Step, Meta
from ..models import Input, Output, ArtifactType, input_to_artifact
from ...project import Stage
from ...utilities.docker import DockerConfig, build, docker_image_tag, docker_file_path

DOCKER_IGNORE_DEFAULT = ['**/target/*', '**/.mpl/*']


class BuildDocker(Step):

    def __init__(self, logger: Logger) -> None:
        super().__init__(logger=logger, meta=Meta(
            name='Docker Build',
            description='Build docker image',
            version='0.0.1',
            stage=Stage.BUILD
        ), produced_artifact=ArtifactType.DOCKER_IMAGE, required_artifact=ArtifactType.NONE,
                         after=AfterBuildDocker(logger=logger))

    def execute(self, step_input: Input) -> Output:
        docker_config = DockerConfig.from_dict(step_input.run_properties.config)
        build_target = docker_config.build_target
        if not build_target:
            raise ValueError('docker.buildTarget must be specified')

        image_tag = docker_image_tag(step_input)
        dockerfile = docker_file_path(project=step_input.project, docker_config=docker_config)

        success = build(logger=self._logger, root_path=docker_config.root_folder,
                        file_path=dockerfile, image_tag=image_tag,
                        target=build_target)
        artifact = input_to_artifact(ArtifactType.DOCKER_IMAGE, step_input, spec={'image': image_tag})

        with open('.dockerignore', 'w+', encoding='utf-8') as ignore_file:
            contents = '\n'.join(DOCKER_IGNORE_DEFAULT)
            ignore_file.write(contents)

        if success:
            return Output(success=True, message=f"Built {step_input.project.name}", produced_artifact=artifact)

        return Output(success=False, message=f"Failed to build docker image for {step_input.project.name}",
                      produced_artifact=None)



"""
This module contains the sealed secret CRD.
"""

from kubernetes.client import V1ObjectMeta

from . import CustomResourceDefinition


class V1SealedSecret(CustomResourceDefinition):
    def __init__(self, name: str, secrets: dict[str, str]):
        super().__init__(api_version="bitnami.com/v1alpha1", kind="SealedSecret",
                         metadata=V1ObjectMeta(name=name, labels={'chart': 'service-0.1.0'},
                                               annotations={'sealedsecrets.bitnami.com/cluster-wide': 'true'}),
                         spec={'encryptedData': secrets})



""" A step to deploy a job to kubernetes. """

from logging import Logger

from .k8s import deploy_helm_chart
from .k8s.chart import ChartBuilder, to_cron_job_chart, to_job_chart
from .. import Step, Meta
from ..models import Input, Output, ArtifactType
from ...project import Stage


class DeployKubernetesJob(Step):

    def __init__(self, logger: Logger) -> None:
        super().__init__(logger, Meta(
            name='Kubernetes Job Deploy',
            description='Deploy a job to k8s',
            version='0.0.1',
            stage=Stage.DEPLOY
        ), produced_artifact=ArtifactType.NONE, required_artifact=ArtifactType.DOCKER_IMAGE)

    def execute(self, step_input: Input) -> Output:
        builder = ChartBuilder(step_input)
        chart = to_cron_job_chart(builder) if builder.is_cron_job else to_job_chart(builder)
        return deploy_helm_chart(self._logger, chart, step_input, builder.release_name, delete_existing=True)



""" Discovery of projects that are relevant to a specific `mpyl.stage.Stage` . Determine which of the
discovered projects have been invalidated due to changes in the source code since the last build of the project's
output artifact."""
import operator

from ..project import Project
from ..project import Stage
from ..steps.models import Output
from ..utilities.repo import Revision


def is_invalidated(project: Project, stage: Stage, path: str) -> bool:
    deps = project.dependencies
    deps_for_stage = deps.set_for_stage(stage) if deps else {}

    touched_dependency = next(filter(path.startswith, deps_for_stage), None) if deps else None
    startswith: bool = path.startswith(project.root_path)
    return startswith or touched_dependency is not None


def _to_relevant_changes(project: Project, stage: Stage, change_history: list[Revision]) -> set[str]:
    output: Output = Output.try_read(project.target_path, stage)
    relevant = set()
    for history in reversed(sorted(change_history, key=lambda c: c.ord)):
        if stage == Stage.DEPLOY or output is None or output.produced_artifact is None \
                or output.produced_artifact.revision != history.hash:
            relevant.update(history.files_touched)
        else:
            return relevant

    return relevant


def are_invalidated(project: Project, stage: Stage, change_history: list[Revision]) -> bool:
    if project.stages.for_stage(stage) is None:
        return False

    relevant_changes = _to_relevant_changes(project, stage, change_history)
    return len(set(filter(lambda c: is_invalidated(project, stage, c), relevant_changes))) > 0


def find_invalidated_projects_for_stage(all_projects: set[Project], stage: Stage,
                                        change_history: list[Revision]) -> set[Project]:
    return set(filter(lambda p: are_invalidated(p, stage, change_history), all_projects))


def find_invalidated_projects_per_stage(all_projects: set[Project], change_history: list[Revision]) \
        -> dict[Stage, set[Project]]:
    projects_for_stage = {}
    for stage in Stage:
        projects = find_invalidated_projects_for_stage(all_projects, stage, change_history)
        if projects:
            projects_for_stage[stage] = set(sorted(projects, key=operator.attrgetter('name')))
    return projects_for_stage


def for_stage(projects: set[Project], stage: Stage) -> set[Project]:
    return set(filter(lambda p: p.stages.for_stage(stage), projects))



"""Docker related utility methods"""
import logging
from dataclasses import dataclass
from logging import Logger
from typing import Dict, Optional

from python_on_whales import docker

from ...project import Project
from ...steps.models import Input


@dataclass(frozen=True)
class DockerComposeConfig:
    period_seconds: int
    failure_threshold: int

    @property
    def total_duration(self):
        return self.period_seconds * self.failure_threshold

    @staticmethod
    def from_yaml(config: dict):
        compose_config = config.get('docker', {}).get('compose')
        if not compose_config:
            raise KeyError('docker.compose needs to be defined')
        return DockerComposeConfig(period_seconds=int(compose_config['periodSeconds']),
                                   failure_threshold=int(compose_config['failureThreshold']))


@dataclass(frozen=True)
class DockerConfig:
    host_name: str
    user_name: str
    password: str
    root_folder: str
    build_target: Optional[str]
    test_target: Optional[str]
    docker_file_name: str

    @staticmethod
    def from_dict(config: Dict):
        try:
            registry: Dict = config['docker']['registry']
            build_config: Dict = config['docker']['build']
            return DockerConfig(
                host_name=registry['hostName'],
                user_name=registry['userName'],
                password=registry['password'],
                root_folder=build_config['rootFolder'],
                build_target=build_config.get('buildTarget', None),
                test_target=build_config.get('testTarget', None),
                docker_file_name=build_config['dockerFileName']
            )
        except KeyError as exc:
            raise KeyError(f'Docker config could not be loaded from {config}') from exc


def stream_docker_logging(logger: Logger, generator, task_name: str, level=logging.INFO) -> None:
    while True:
        try:
            output = next(generator)
            logger.log(level, str(output).strip('\n'))
        except StopIteration:
            logger.info(f'{task_name} complete.')
            break


def docker_image_tag(step_input: Input):
    git = step_input.run_properties.versioning
    tag = f"pr-{git.pr_number}" if git.pr_number else git.tag
    return f"{step_input.project.name.lower()}:{tag}".replace('/', '_')


def docker_file_path(project: Project, docker_config: DockerConfig):
    return f'{project.deployment_path}/{docker_config.docker_file_name}'


def build(logger: Logger, root_path: str, file_path: str, image_tag: str, target: str) -> bool:
    """
    :param logger: the logger
    :param root_path: the root path to which `docker_file_path` is relative
    :param file_path: path to the docker file to be built
    :param image_tag: the tag of the image
    :param target: the 'target' within the multi-stage docker image
    :return: True if success, False if failure
    """
    logger.info(f"Building docker image with {file_path} and target {target}")

    logs = docker.buildx.build(context_path=root_path, file=file_path, tags=[image_tag], target=target,
                               stream_logs=True)
    stream_docker_logging(logger=logger, generator=logs, task_name=f'Build {file_path}:{target}')
    logger.debug(logs)
    return True



""" Entry point of MPyL. Loads all available Step implementations and triggers their execution based on the specified
Project and Stage.
"""
import itertools
import pkgutil
from dataclasses import dataclass
from datetime import datetime
from logging import Logger
from typing import Optional
from unittest import TestSuite

from ruamel.yaml import YAML  # type: ignore

from . import Step
from .build.dockerbuild import BuildDocker
from .build.echo import BuildEcho
from .build.sbt import BuildSbt
from .deploy.echo import DeployEcho
from .deploy.ephemeral_docker_deploy import EphemeralDockerDeploy
from .deploy.kubernetes import DeployKubernetes
from .deploy.kubernetes_job import DeployKubernetesJob
from .deploy.kubernetes_spark_job import DeployKubernetesSparkJob
from .models import Output, Input, RunProperties, ArtifactType, Artifact
from .test.dockertest import TestDocker
from .test.echo import TestEcho
from .test.sbt import TestSbt
from ..project import Project
from ..project import Stage
from ..utilities.junit import to_test_suites
from ..validation import validate

yaml = YAML()


class ExecutionException(Exception):
    """ Exception thrown when a step execution fails. """

    def __init__(self, project_name: str, executor: str, stage: str, message: str):
        self.project_name = project_name
        self.executor = executor
        self.stage = stage
        self.message = message
        super().__init__(self.message)


@dataclass(frozen=True)
class StepResult:
    stage: Stage
    project: Project
    output: Output
    timestamp: datetime = datetime.now()


def collect_test_results(step_results: list[StepResult]) -> list[TestSuite]:
    test_artifacts = [res.output.produced_artifact for res in step_results if
                      (res.output.produced_artifact and
                       res.output.produced_artifact.artifact_type == ArtifactType.JUNIT_TESTS)]

    suites: list[list[TestSuite]] = list(map(to_test_suites, test_artifacts))
    return list(itertools.chain(*suites))


class Steps:
    """ Executor of individual steps within a pipeline. """
    _step_executors: dict[Stage, set[Step]]
    _logger: Logger
    _properties: RunProperties

    def __init__(self, logger: Logger, properties: RunProperties) -> None:
        schema_dict = pkgutil.get_data(__name__, "../schema/mpyl_config.schema.yml")

        if schema_dict:
            validate(properties.config, schema_dict.decode('utf-8'))

        self._logger = logger

        self._step_executors: dict[Stage, set[Step]] = {
            Stage.BUILD: {
                BuildEcho(logger),
                BuildSbt(logger),
                BuildDocker(logger)
            },
            Stage.TEST: {
                TestEcho(logger),
                TestSbt(logger),
                TestDocker(logger)
            },
            Stage.DEPLOY: {
                DeployEcho(logger),
                DeployKubernetes(logger),
                DeployKubernetesJob(logger),
                DeployKubernetesSparkJob(logger),
                EphemeralDockerDeploy(logger)
            }
        }

        self._properties = properties
        for stage, steps in self._step_executors.items():
            self._logger.debug(f"Registered executors for stage {stage.name}: "  # pylint: disable=E1101
                               f"{[step.meta.name for step in steps]}")

    def _find_executor(self, stage: Stage, step_name: str) -> Optional[Step]:
        executors = filter(lambda e: e.meta.stage == stage and step_name == e.meta.name, self._step_executors[stage])
        return next(executors, None)

    def _execute(self, executor: Step, project: Project, properties: RunProperties,
                 artifact: Optional[Artifact], dry_run: bool = False) -> Output:
        result = executor.execute(Input(project, properties, required_artifact=artifact, dry_run=dry_run))
        if result.success:
            self._logger.info(
                f"Execution of {executor.meta.name} succeeded for '{project.name}' with outcome '{result.message}'")
        else:
            self._logger.warning(
                f"Execution of {executor.meta.name} failed for '{project.name}' with outcome '{result.message}'")
        return result

    @staticmethod
    def _find_required_artifact(project: Project, required_artifact: Optional[ArtifactType]) -> Optional[Artifact]:
        if not required_artifact or required_artifact == ArtifactType.NONE:
            return None

        for stage in Stage:
            output: Optional[Output] = Output.try_read(project.target_path, stage)
            if output and output.produced_artifact and output.produced_artifact.artifact_type == required_artifact:
                return output.produced_artifact

        raise ValueError(f"Artifact {required_artifact} required for {project.name} not found")

    def _execute_stage(self, stage: Stage, project: Project, dry_run: bool = False) -> Output:
        stage_name = project.stages.for_stage(stage)
        if stage_name is None:
            return Output(success=False, message=f"Stage '{stage.value}' not defined on project '{project.name}'")

        executor = self._find_executor(stage, stage_name)
        if executor:
            try:
                self._logger.info(f'Executing {stage} for {project.name}')
                artifact: Optional[Artifact] = self._find_required_artifact(project, executor.required_artifact)
                result = Output(success=True, message='')
                if executor.before:
                    result = self._execute(executor.before, project, self._properties,
                                           self._find_required_artifact(project, executor.before.required_artifact),
                                           dry_run)
                if result.success:
                    result = self._execute(executor, project, self._properties, artifact, dry_run)
                    result.write(project.target_path, stage)
                if executor.after:
                    main_step_artifact = result.produced_artifact
                    result = self._execute(executor.after, project, self._properties, result.produced_artifact, dry_run)
                    if result.produced_artifact and result.produced_artifact.artifact_type != ArtifactType.NONE:
                        result.write(project.target_path, stage)
                    else:
                        result.produced_artifact = main_step_artifact

                return result
            except Exception as exc:
                message = str(exc)
                self._logger.warning(
                    f"Execution of '{executor.meta.name}' for project '{project.name}' in stage {stage} "
                    f"failed with exception: {message}", exc_info=True)
                raise ExecutionException(project.name, executor.meta.name, stage.name, message) from exc
        else:
            self._logger.warning(f"No executor found for {stage_name} in stage {stage}")

        return Output(success=False, message=f"Executor '{stage_name}' for '{stage.value}' not known or registered")

    def execute(self, stage: Stage, project: Project, dry_run: bool = False) -> StepResult:
        """
        :param stage: the stage to execute
        :param project: the project metadata
        :param dry_run: indicates whether artifacts should be submitted or deployed for real
        :return: StepResult
        :raise ExecutionException
        """
        step_output = self._execute_stage(stage, project, dry_run)
        return StepResult(stage=stage, project=project, output=step_output)



"""Command Line Interface parsing for MPyL"""
import asyncio
import importlib
import logging
from dataclasses import dataclass
from importlib.metadata import version as version_meta
from typing import Optional

import aiohttp
import requests
from aiohttp import ClientConnectorError, ClientTimeout

from rich.console import Console
from rich.logging import RichHandler

from ..utilities.repo import Repository

CONFIG_PATH_HELP = 'Path to the config.yml. Needs to comply with schema at ' \
                   'https://vandebron.github.io/mpyl/schema/mpyl_config.schema.yml ' \
                   'Can be set via `MPYL_CONFIG_PATH` env var. '


@dataclass(frozen=True)
class CliContext:
    config: dict
    repo: Repository
    console: Console
    verbose: bool


async def fetch_latest_version() -> Optional[str]:
    try:
        async with aiohttp.ClientSession(timeout=ClientTimeout(total=10)) as session:
            async with session.get("https://pypi.org/pypi/mpyl/json") as response:
                body = await response.json()
                return body.get('info', {}).get('version')
    except (asyncio.exceptions.TimeoutError, ClientConnectorError, requests.exceptions.RequestException):
        return None


def get_meta_version():
    try:
        return version_meta('mpyl')
    except importlib.metadata.PackageNotFoundError:
        return None


async def check_updates(meta: str) -> Optional[str]:
    latest = await fetch_latest_version()
    if latest and meta != latest:
        return latest
    return None


def get_version():
    try:
        return f"v{version_meta('mpyl')}"
    except importlib.metadata.PackageNotFoundError:
        return '(local)'


FORMAT = "%(message)s"


def create_console_logger(local: bool, verbose: bool) -> Console:
    console = Console(markup=True, width=None if local else 135, no_color=False, log_path=False, log_time=False,
                      color_system='256')
    logging.basicConfig(
        level="DEBUG" if verbose else "INFO", format=FORMAT, datefmt="[%X]",
        handlers=[RichHandler(markup=True, console=console, show_path=local)]
    )
    return console



"""Health checks"""

import asyncio
import os
import pkgutil
from pathlib import Path
from subprocess import CalledProcessError

import jsonschema
from dotenv import load_dotenv
from rich.console import Console
from rich.markdown import Markdown

from ..build.jenkins import get_token
from ....utilities.github import GithubConfig
from ....utilities.jenkins import JenkinsConfig
from ....utilities.pyaml_env import parse_config
from ....cli import fetch_latest_version, get_meta_version
from ....validation import validate


def perform_health_checks(console: Console):
    console.print(Markdown('*Version*'))
    __check_version(console)

    console.print(Markdown('*MPyL configuration*'))
    __check_config(console, env_var='MPYL_CONFIG_PATH', default='mpyl_config.yml',
                   schema_path='../../../schema/mpyl_config.schema.yml', name='config')

    console.print(Markdown('*Run configuration*'))
    __check_config(console, env_var='MPYL_RUN_PROPERTIES_PATH', default='run_properties.yml',
                   schema_path='../../../schema/run_properties.schema.yml', name='run properties')

    console.print(Markdown('*Jenkins*'))
    __check_jenkins(console)


def __check_jenkins(console):
    path = os.environ.get('MPYL_CONFIG_PATH', default='mpyl_config.yml')
    if not os.path.exists(path):
        return

    parsed = parse_config(path)

    try:
        jenkins_conf = JenkinsConfig.from_config(parsed)
        console.log(f'  ✅ Jenkins configured for pipeline `{jenkins_conf.default_pipeline}` at {jenkins_conf.url}')
    except KeyError as exc:
        console.log(f'  ❌ Jenkins config not valid: {exc}')

    try:
        get_token(GithubConfig(parsed))
        console.log('  ✅ Github token found')
    except CalledProcessError:
        console.log('  ❌ Github token not found. Install Github CLI `brew install gh`')

    if os.environ.get('JENKINS_USER'):
        console.log('  ✅ Jenkins user set')
    else:
        console.log("  ❌ Jenkins user not set via JENKINS_USER env var")
        jenkins_url = f'{JenkinsConfig.from_config(parsed).url}user/me@vandebron.nl/configure'
        console.log(f"     Create a user API token in Jenkins (user:password) API token: {jenkins_url}")

    if os.environ.get('JENKINS_PASSWORD'):
        console.log('  ✅ Jenkins password set')
    else:
        console.log("  ❌ Jenkins password not set via JENKINS_PASSWORD env var")


def __check_version(console):
    update = asyncio.get_event_loop().run_until_complete(fetch_latest_version())
    meta_version = get_meta_version()
    if update and meta_version:
        if meta_version == update:
            console.log(f'  ✅ At latest version: {update}')
        else:
            console.log(f'  ❌ Outdated version: {meta_version} (latest: {update})')
    else:
        console.log('  ❌ Could not determine latest version')


def __check_config(console, env_var, default, schema_path, name):
    path = os.environ.get(env_var, default=default)
    location = f"{name} at '/{path}' via environment variable '{env_var}'" if os.environ.get(
        env_var) else f"{name} at '/{path}'"
    if os.path.exists(path):
        console.log(f"  ✅ Found {location}")

        if load_dotenv(Path(".env")):
            console.log("  ✅ Set env variables via .env file")

        parsed = parse_config(path)
        schema_dict = pkgutil.get_data(__name__, schema_path)
        if schema_dict:
            try:
                validate(parsed, schema_dict.decode('utf-8'))
                console.log(f'  ✅ {name.capitalize()} is valid')
            except jsonschema.exceptions.ValidationError as exc:
                console.log(f"  ❌ {name.capitalize()} is invalid: {exc.message} at '{'.'.join(exc.path)}'")
    else:
        console.log(f"  ❌ Could not find {location}. Location can be specified with env var '{env_var}'")



"""
Wrapper around `pyaml_env`'s `parse_config`. Sets default values in config to `None`
"""
from pathlib import Path

from pyaml_env import parse_config as original_parse_config
from dotenv import load_dotenv


def parse_config(path: str) -> dict[str, str]:
    load_dotenv(Path(".env"))

    def default_to_none(obj, default_value='N/A'):
        if isinstance(obj, (list, tuple, set)):
            return type(obj)(default_to_none(x) for x in obj if x is not default_value)
        if isinstance(obj, dict):
            return type(obj)((default_to_none(k), default_to_none(v))
                             for k, v in obj.items() if k is not default_value and v is not default_value)
        if obj == default_value:
            return None
        return obj

    parsed = original_parse_config(path)
    return default_to_none(parsed)



"""Entrypoint for cli"""

if __name__ == '__main__' and __package__ is None:
    # https://peps.python.org/pep-0366/
    __package__ = "mpyl"  # pylint: disable=redefined-builtin

    from . import main

    main()



"""
This module contains the traefik ingress route CRD.
"""
from dataclasses import dataclass
from typing import Optional

from kubernetes.client import V1ObjectMeta

from . import CustomResourceDefinition
from .....project import Host, Target


@dataclass(frozen=True)
class HostWrapper:
    host: Host
    name: str
    index: int
    service_port: int
    white_lists: list[str]

    @property
    def full_name(self) -> str:
        return f'{self.name}-ingress-{self.index}-whitelist'


class V1AlphaIngressRoute(CustomResourceDefinition):

    def __init__(self, metadata: V1ObjectMeta, hosts: list[HostWrapper], target: Target,
                 pr_number: Optional[int]):
        def _interpolate_names(host: str, name: str) -> str:
            host = host.replace('{SERVICE-NAME}', name)
            if pr_number:
                return host.replace('{PR-NUMBER}', str(pr_number))
            return host

        routes = [{'kind': 'Rule', 'match': _interpolate_names(host.host.host.get_value(target), host.name),
                   'services': [{'name': host.name, 'kind': 'Service', 'port': host.service_port}],
                   'middlewares': [{'name': host.full_name}]} for host in hosts]

        super().__init__(api_version='traefik.containo.us/v1alpha1', kind="IngressRoute", metadata=metadata,
                         spec={'routes': routes, 'entryPoints': ['websecure'],
                               'tls': {'secretName': 'le-prod-wildcard-cert'}}, schema='traefik.ingress.schema.yml')


class V1AlphaMiddleware(CustomResourceDefinition):

    def __init__(self, metadata: V1ObjectMeta, source_ranges: list[str]):
        super().__init__(api_version='traefik.containo.us/v1alpha1', kind="Middleware", metadata=metadata,
                         spec={'ipWhiteList': {'sourceRange': source_ranges}}, schema='traefik.middleware.schema.yml')



""" Before test step. Starts docker-compose if necessary."""
import os
import time
from logging import Logger

from python_on_whales import DockerClient, Container
from python_on_whales.components.compose.models import ComposeProject

from .. import Step, Meta
from ..models import Input, Output, ArtifactType
from ...project import Stage
from ...utilities.docker import stream_docker_logging, DockerComposeConfig


class IntegrationTestBefore(Step):
    def __init__(self, logger: Logger) -> None:
        super().__init__(logger=logger, meta=Meta(
            name='Before Test',
            description='Before test step',
            version='0.0.1',
            stage=Stage.TEST
        ), produced_artifact=ArtifactType.NONE, required_artifact=ArtifactType.NONE)

    def execute(self, step_input: Input) -> Output:
        compose_file = step_input.project.test_containers_path
        if not os.path.exists(compose_file):
            return Output(success=True, message='No containers to start')

        config = DockerComposeConfig.from_yaml(step_input.run_properties.config)

        self._logger.debug(f"Starting containers in {compose_file}")
        docker_client = DockerClient(compose_files=[compose_file])
        docker_client.compose.down(remove_orphans=True)
        docker_client.compose.build()
        docker_client.compose.up(detach=True, color=True, quiet=False)

        goal_reached = False
        logs = docker_client.compose.logs(stream=True)
        stream_docker_logging(logger=self._logger, generator=logs, task_name=f'Start {compose_file}')

        poll = 0
        while not goal_reached:
            proj: ComposeProject = docker_client.compose.ls()[0]
            goal_reached = (proj.created + proj.restarting + proj.exited + proj.paused + proj.dead) == 0
            if not goal_reached:
                self._logger.debug(f'Project stats: {proj}')

            poll += 1
            if poll >= config.failure_threshold:
                return Output(success=False,
                              message=f"Failed to start services in {compose_file} "
                                      f"within {config.total_duration} seconds.")

            time.sleep(config.period_seconds)
        running_containers: list[Container] = docker_client.compose.ps()
        container_names = list(map(lambda l: l.name, running_containers))

        return Output(success=True, message=f"Started {', '.join(container_names)}")



"""
Interface for custom reporters.
To create a new reporter, implement `Reporter.send_report`.
You can transform `mpyl.steps.run.RunResult` to a report using any of the standard formatters under
`mpyl.reporting.formatting` and send it to target of your choice.

See the **Submodules** for built-in reporters.
"""
from abc import abstractmethod
from typing import Optional

from ...steps.run import RunResult


class Reporter:

    @abstractmethod
    def send_report(self, results: RunResult, text: Optional[str] = None) -> None:
        pass



"""Serialization of projects for display purposes"""

from pathlib import Path

from rich.console import Console
from rich.table import Table

from ... import Repository
from ....project import load_project
from ....projects.find import load_projects, find_dependencies


def print_project(repo: Repository, console: Console, project_path: str):
    project = load_project(repo.root_dir(), Path(project_path), False)
    other_projects = load_projects(repo.root_dir(), repo.find_projects())

    with_dependencies = find_dependencies(project, other_projects)

    table = Table(title=f"Project {project.name}", show_header=False)
    table.add_row("Name", project.name)
    table.add_row("Path", project.path)
    table.add_row("Description", project.description)
    table.add_row("Maintainer", f"{project.maintainer}")
    table.add_row("Stages", f"{project.stages}")
    if with_dependencies.dependent_projects:
        table.add_row("Dependent projects", f"{set(with_dependencies.dependent_projects.keys())}")
    console.print(table)



"""
Simple run result formatters
"""

from ...project import Stage
from ...steps.models import Artifact, ArtifactType
from ...steps.run import RunResult
from ...utilities.junit import to_test_suites, sum_suites


def to_string(run_result: RunResult) -> str:
    result: str = ""
    for stage in Stage:
        run_results = run_result.results_for_stage(stage)
        if run_results:
            result += f"Stage {stage.name}\n"
            for res in run_results:
                result += f"{res.timestamp} - {res.project.name} - {res.stage} - success: {res.output.success} \n"
                artifact = res.output.produced_artifact
                if artifact and artifact.artifact_type == ArtifactType.JUNIT_TESTS:
                    result += to_test_report(artifact)

    return result


def to_test_report(artifact: Artifact):
    test_result = ""
    suites = to_test_suites(artifact)
    total_tests = sum_suites(suites)
    test_result += f"{total_tests} \n\n"
    for suite in suites:
        test_result += f"Suite {suite.name}\n"
        for case in suite:
            test_result += f"Case {case.name} \n"

    return test_result



""" Function used to validate the project.schema.yml against the local schema."""

import pkgutil
from functools import lru_cache

import jsonschema
from jsonschema import RefResolver, Draft7Validator, validators
from ruamel.yaml import YAML

yaml = YAML()


@lru_cache(maxsize=10)
def load_schema(schema_string: str) -> Draft7Validator:
    schema = yaml.load(schema_string)
    project_schema_string = pkgutil.get_data(__name__, "schema/project.schema.yml")
    if not project_schema_string:
        raise ImportError("'schema/project.schema.yml' was not found in bundle")

    project_schema = yaml.load(project_schema_string.decode('utf-8'))

    def load_schema_from_local(uri):
        if 'project.schema.yml' in uri:
            return project_schema
        return {}

    resolver = RefResolver(referrer=schema, base_uri="", handlers={"": load_schema_from_local})
    all_validators = dict(Draft7Validator.VALIDATORS)
    existing_validator = all_validators['type']

    def allow_none_validator(validator, types, instance, yaml_schema):
        for field_type in types:
            if field_type is None and instance is None:
                return None

        return existing_validator(validator, types, instance, yaml_schema)

    all_validators['type'] = allow_none_validator
    type_checker = Draft7Validator.TYPE_CHECKER

    extended_validator = validators.extend(jsonschema.validators.Draft7Validator, validators=all_validators,
                                           type_checker=type_checker)
    return extended_validator(schema=schema, resolver=resolver)


def validate(values: dict, schema_string: str):
    schema = load_schema(schema_string)
    return schema.validate(values)



"""Kubernetes deployment related helper methods"""
from logging import Logger

from kubernetes import config, client

from ...deploy.k8s.resources import CustomResourceDefinition
from ...models import RunProperties
from ....project import Project, Target
from ....steps import Input, Output
from ....steps.deploy.k8s import helm
from ....steps.deploy.k8s.rancher import cluster_config, rancher_namespace_metadata


def get_namespace(run_properties: RunProperties, project: Project) -> str:
    if run_properties.target == Target.PULL_REQUEST:
        return run_properties.versioning.identifier

    if project.deployment and project.deployment.namespace:
        return project.deployment.namespace

    return project.name


def upsert_namespace(logger: Logger, step_input: Input, context: str):
    properties = step_input.run_properties

    config.load_kube_config(context=context)
    logger.info(f"Deploying target {properties.target} and k8s context {context}")
    api = client.CoreV1Api()

    namespace = get_namespace(properties, step_input.project)
    meta_data = rancher_namespace_metadata(namespace, step_input)
    namespaces = api.list_namespace(field_selector=f'metadata.name={namespace}')

    if len(namespaces.items) == 0 and not step_input.dry_run:
        api.create_namespace(client.V1Namespace(api_version='v1', kind='Namespace', metadata=meta_data))
    else:
        logger.info(f"Found namespace {namespace}")

    return namespace


def deploy_helm_chart(logger: Logger, chart: dict[str, CustomResourceDefinition], step_input: Input,
                      release_name: str, delete_existing: bool = False) -> Output:
    context = cluster_config(step_input).context
    namespace = upsert_namespace(logger, step_input, context)

    return helm.install(logger, chart, step_input, release_name, namespace, context, delete_existing)



"""
Jira result reporter.

### Usage

To determine which `Jira` ticket a pipeline run relates to, the branch is parsed.
The following pattern is assumed unless otherwise specified: `[A-Za-z]{3,}-` to signify the Jira ticket key.
If a Jira ticket key cannot be extracted from the branch, this reporter will fail silently.

### Actions
 - If the ticket is still in its initial stage, it transitioned to the following stage. For example:
`To Do` -> `In progress`
 - If the ticket hasn't been assigned to any particular person, it will be assigned to whoever triggered
 the pipeline run.

### Configuration
Configure hostname and credentials under `jira` in the `config.yml`

.. note:: Use API token as password
   You need to create a regular Jira account to represent the MPyL system integration.
   The username is the email address of the account. Log in as this user and create an
   [API token](https://id.atlassian.com/manage-profile/security/api-tokens) to be used as the password.
"""
import re
from dataclasses import dataclass
from logging import Logger
from re import Pattern
from typing import Optional
from urllib.parse import urlsplit

import requests
from atlassian import Jira

from . import Reporter
from ..formatting.markdown import markdown_for_stage
from ...project import Stage
from ...steps.run import RunResult


@dataclass(frozen=True)
class JiraTicket:
    jira_url: str
    ticket_id: str
    ticket_url: str
    issue_type: str
    summary: str
    description: str
    status_name: str
    user_avatar: str
    user_email: str
    assignee_email: Optional[str]

    @staticmethod
    def from_issue_response(response):
        jira_url = "{uri.scheme}://{uri.netloc}".format(uri=urlsplit(response['self']))
        fields = response['fields']
        ticket_id = response['key']
        assignee = fields.get('assignee')
        user = assignee or fields.get('reporter') or fields.get('creator')
        avatar = user['avatarUrls']['24x24'] if user else ''
        status_name = fields['status']['name']
        assignee_email = assignee['emailAddress'] if assignee else None
        ticket_url = f"{jira_url}/browse/{ticket_id}"

        return JiraTicket(jira_url=jira_url, ticket_id=ticket_id, ticket_url=ticket_url,
                          issue_type=fields['issuetype']['name'], summary=fields['summary'], status_name=status_name,
                          description=fields['description'], user_avatar=avatar, user_email=user['emailAddress'],
                          assignee_email=assignee_email)


def extract_ticket_from_branch(branch: str, pattern: Pattern) -> Optional[str]:
    ticket: Optional[str] = next(iter(re.findall(pattern, branch)), None)
    if ticket:
        return ticket.upper()
    return None


@dataclass(frozen=True)
class JiraConfig:
    site: str
    user_name: str
    password: str
    ticket_pattern: Pattern
    token: Optional[str]

    @staticmethod
    def from_config(config: dict):
        jira_config = config.get('jira')
        if not jira_config:
            raise KeyError('jira section needs to be defined in mpyl_config.yml')
        return JiraConfig(site=jira_config['site'], user_name=jira_config['userName'], password=jira_config['password'],
                          ticket_pattern=re.compile(jira_config.get('ticketPattern', '[A-Za-z]{3,}-\\d+')),
                          token=jira_config.get('token'))


def create_jira_for_config(jira_config: JiraConfig):
    return Jira(
        url=jira_config.site,
        token=jira_config.token,
        api_version='2',
        cloud=True
    ) if jira_config.token else Jira(
        url=jira_config.site,
        username=jira_config.user_name,
        password=jira_config.password,
        api_version='2',
        cloud=True
    )


def to_github_markdown(jira_markdown: str, jira_url: str) -> str:
    jira_markdown = re.sub(r'\[(.*)/(.*)\|(.*)\|smart-link\]', r'[\2](\3)', jira_markdown)
    jira_markdown = re.sub(r'\[(.*)\|(.*)\]', r'[\1](\2)', jira_markdown)
    jira_markdown = re.sub(r'\[~accountid:(.*)\]', rf'[👩‍💻]({jira_url}/jira/people/\1)', jira_markdown)
    jira_markdown = re.sub(r'\{\{(.*)\}\}', r'`\1`', jira_markdown)
    jira_markdown = re.sub(r'\{quote}(.*)\{quote}', r'> \1', jira_markdown)
    jira_markdown = re.sub(r'\{noformat\}((.|\n)*)\{noformat\}', r'```\n\1\n```', jira_markdown)
    jira_markdown = re.sub(r'\{noformat\}((.|\n)*)', r'```\n\1\n```', jira_markdown)
    jira_markdown = re.sub(r'\*(.*)\*', r'**\1**', jira_markdown)
    jira_markdown = re.sub(r'_(.*)_', r'*\1*', jira_markdown)
    jira_markdown = re.sub(r'!.*\|.*!', r'', jira_markdown)

    jira_markdown = jira_markdown.replace('h1. ', '### ')
    jira_markdown = jira_markdown.replace('h2. ', '#### ')
    jira_markdown = jira_markdown.replace('h3. ', '##### ')
    jira_markdown = jira_markdown.replace('h4. ', '###### ')
    jira_markdown = jira_markdown.replace('h5. ', '###### ')
    jira_markdown = jira_markdown.replace('h6. ', '###### ')
    return jira_markdown


def to_markdown_summary(ticket: JiraTicket, run_result: RunResult) -> str:
    description_markdown = to_github_markdown(ticket.description, ticket.ticket_url) if ticket.description else ""
    lines = description_markdown.splitlines()
    max_message_length = 288
    if len(lines) > max_message_length:
        description_markdown = "\n".join(lines[:max_message_length]) + "\n..."

    details = run_result.run_properties.details

    build_status = f"🏗️ Build [{details.build_id}]({details.run_url}) {run_result.status_line}, " \
                   f"started by _{details.user}_  \n{markdown_for_stage(run_result, Stage.DEPLOY)}"
    return f"## 📕 [{ticket.ticket_id}]({ticket.ticket_url}) {ticket.summary} " \
           f"![{ticket.user_email}]({ticket.user_avatar}) \n" \
           f"{description_markdown}\n" \
           f"{build_status}"


def compose_build_status(result: RunResult, config: dict) -> str:
    jira_config = JiraConfig.from_config(config=config)
    jira_client = create_jira_for_config(jira_config)
    branch = result.run_properties.versioning.branch
    if not branch:
        return " # ⚠️ `versioning.branch` not set, cannot find corresponding ticket"
    ticket_id = extract_ticket_from_branch(branch, jira_config.ticket_pattern)
    if not ticket_id:
        return f" # ⚠️ Could not find ticket corresponding to `{branch}. " \
               f"Does your branch name follow the correct pattern?"
    issue = jira_client.get_issue(ticket_id)
    jira_ticket = JiraTicket.from_issue_response(issue)
    return to_markdown_summary(jira_ticket, result)


class JiraReporter(Reporter):

    def __init__(self, config: dict, branch: str, logger: Logger):
        jira_config = JiraConfig.from_config(config)
        self._ticket = extract_ticket_from_branch(branch, jira_config.ticket_pattern)
        self._config = jira_config
        self._jira = create_jira_for_config(jira_config)
        self._logger = logger

    def send_report(self, results: RunResult, text: Optional[str] = None) -> None:
        if not self._ticket:
            return None

        try:
            issue_response = self._jira.get_issue(self._ticket)
        except requests.exceptions.HTTPError as exc:
            self._logger.warning(f'Could not find JIRA ticket {self._ticket}: {exc}')
            raise exc

        ticket = JiraTicket.from_issue_response(issue_response)

        self.__move_ticket_forward(ticket)

        user_email = results.run_properties.details.user_email
        if user_email:
            self.__assign_ticket(user_email, ticket)
        return None

    def __assign_ticket(self, run_user_email: str, ticket: JiraTicket):
        if ticket.assignee_email is None or ticket.assignee_email != run_user_email:
            jira_user = self._jira.user_find_by_user_string(query=run_user_email)
            if jira_user:
                account_id = jira_user.pop().get('accountId')
                self._logger.info(f'Assigning {ticket.ticket_id} to {account_id}')
                self._jira.assign_issue(self._ticket, account_id=account_id)

    def __move_ticket_forward(self, ticket: JiraTicket):
        transitions = self._jira.get_issue_transitions(self._ticket)
        for idx, transition in enumerate(transitions):
            if transition['name'] == ticket.status_name:
                if idx == 0:
                    target_state = transitions[idx + 1]['name']
                    self._logger.info(f'Moving {ticket.ticket_id} to {target_state}')
                    self._jira.issue_transition(self._ticket, target_state)



""" Defines information about the repository, any changes made to it and the containing projects.
`mpyl.utilities.repo.Repository` is a facade for the Version Control System.
At this moment Git is the only supported VCS.
"""

from dataclasses import dataclass
from pathlib import Path
from typing import Dict

from git import Git, Repo, Remote

from ...project import Project


@dataclass(frozen=True)
class Revision:
    ord: int
    """Ordinal number indicating how this revision ranks historically"""
    hash: str
    """Git hash for this revision"""
    files_touched: set[str]
    """Paths to files that were altered in this hash"""


class RepoConfig:
    main_branch: str

    def __init__(self, config: Dict):
        self.main_branch = config['cvs']['git']['mainBranch']


class Repository:

    def __init__(self, config: RepoConfig):
        self._config = config
        self._root_dir = Git().rev_parse('--show-toplevel')
        self._repo = Repo(self._root_dir)  # pylint: disable=attribute-defined-outside-init

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._repo.close()
        return self

    @property
    def get_sha(self):
        return self._repo.head.commit.hexsha

    @property
    def get_short_sha(self):
        return self._repo.git.rev_parse(self._repo.head, short=True)

    @property
    def get_branch(self):
        return self._repo.active_branch.name

    @property
    def get_remote_url(self):
        return self._repo.remote().url

    def root_dir(self) -> Path:
        return Path(self._root_dir)

    @property
    def main_branch(self) -> str:
        return self._config.main_branch

    @property
    def latest_tag(self) -> str:
        return str(sorted(self._repo.tags, key=lambda t: t.commit.committed_datetime)[-1])

    def changes_in_branch(self) -> list[Revision]:
        revisions = reversed(list(self._repo.iter_commits(f"{self._config.main_branch}..HEAD")))
        return [Revision(count, str(rev),
                         self._repo.git.diff_tree(no_commit_id=True, name_only=True, r=str(rev)).splitlines()) for
                count, rev in enumerate(revisions)]

    def changes_in_branch_including_local(self) -> list[Revision]:
        in_branch = self.changes_in_branch()
        in_branch.append(Revision(len(in_branch), self.get_sha, self.changes_in_commit()))
        return in_branch

    @property
    def main_branch_pulled(self) -> bool:
        branch_names = list(map(lambda n: n.name, self._repo.references))
        return f'origin/{self._config.main_branch}' in branch_names

    def pull_main_branch(self):
        remote = Remote(self._repo, 'origin')
        main = self._config.main_branch
        return remote.fetch(f"+refs/heads/{main}:refs/heads/{main}")

    def changes_in_commit(self) -> set[str]:
        changed: set[str] = set(self._repo.git.diff(None, name_only=True).splitlines())
        return changed.union(self._repo.untracked_files)

    def find_projects(self, folder_pattern: str = '') -> list[str]:
        """ returns a set of all project.yml files
        :type folder_pattern: project paths are filtered on this pattern
        """
        projects = set(self._repo.git.ls_files(f'*{folder_pattern}*/{Project.project_yaml_path()}').splitlines())
        return sorted(projects)



"""Github related utility methods"""
from dataclasses import dataclass
from typing import Dict, Optional

from github.PullRequest import PullRequest
from github.Repository import Repository


@dataclass
class GithubAppConfig:
    private_app_key_path: Optional[str]
    private_key_base_64_encoded: Optional[str]
    app_key: str

    def __init__(self, config: Dict):
        self.private_app_key_path = config.get('privateKeyPath')
        self.private_key_base_64_encoded = config.get('privateKeyBase64Encoded')
        if not self.private_key_base_64_encoded and not self.private_app_key_path:
            raise KeyError("When github.app is configured, either 'privateKeyPath' "
                           "or 'privateKeyBase64Encoded' need to be defined")

        self.app_key = config['appId']


@dataclass
class GithubConfig:
    repository: str
    owner: str
    repo_name: str
    token: str
    app_config: dict

    def __init__(self, config: Dict):
        github = config['cvs']['github']
        self.repository = github['repository']
        parts = self.repository.split('/')
        self.owner = parts[0]
        self.repo_name = parts[1]
        self.token = github['token']
        self.app_config = github.get('app', {})

    @property
    def get_app_config(self) -> GithubAppConfig:
        return GithubAppConfig(self.app_config)


def get_pr_for_branch(repo: Repository, branch: str) -> PullRequest:
    pulls = repo.get_pulls(head=f'{repo.full_name}:{branch}').get_page(0)

    if len(pulls) == 0:
        raise ValueError(f'No PR related to {branch} was found. Did create it yet? `gh pr create --draft`')

    return pulls.pop()



""" Step that tests the docker image from the target `tester` in Dockerfile-mpl. """
import shutil
from logging import Logger
from pathlib import Path

from python_on_whales import docker

from .after_test import IntegrationTestAfter
from .before_test import IntegrationTestBefore
from .. import Step, Meta
from ..models import Input, Output, ArtifactType, input_to_artifact, Artifact
from ...project import Stage, Project
from ...utilities.docker import DockerConfig, build, docker_image_tag, docker_file_path
from ...utilities.junit import to_test_suites, sum_suites, TEST_OUTPUT_PATH_KEY


class TestDocker(Step):
    def __init__(self, logger: Logger) -> None:
        meta = Meta(name='Docker Test', description='Test docker image', version='0.0.1', stage=Stage.TEST)
        super().__init__(
            logger=logger, meta=meta,
            produced_artifact=ArtifactType.JUNIT_TESTS,
            required_artifact=ArtifactType.NONE,
            before=IntegrationTestBefore(logger),
            after=IntegrationTestAfter(logger)
        )

    def execute(self, step_input: Input) -> Output:
        docker_config = DockerConfig.from_dict(step_input.run_properties.config)
        test_target = docker_config.test_target
        if not test_target:
            raise ValueError('docker.testTarget must be specified')

        tag = docker_image_tag(step_input) + '-test'
        project = step_input.project
        dockerfile = docker_file_path(project=project, docker_config=docker_config)

        success = build(logger=self._logger, root_path=docker_config.root_folder,
                        file_path=dockerfile, image_tag=tag, target=test_target)

        if success:
            artifact = self.extract_test_results(project, tag, step_input)

            suite = to_test_suites(artifact)
            summary = sum_suites(suite)

            return Output(success=summary.is_success, message=f"Tests results produced for {project.name} ({summary})",
                          produced_artifact=artifact)

        return Output(success=False,
                      message=f"Tests failed to run for {project.name}. No test results have been recorded.",
                      produced_artifact=None)

    @staticmethod
    def extract_test_results(project: Project, tag, step_input: Input) -> Artifact:
        test_result_path = Path(project.target_path, "test_results")
        shutil.rmtree(test_result_path, ignore_errors=True)
        Path(test_result_path).mkdir(parents=True, exist_ok=True)

        container_id = docker.create(tag).id
        docker.copy(f'{container_id}:/{project.test_report_path}/.', test_result_path)

        return input_to_artifact(artifact_type=ArtifactType.JUNIT_TESTS, step_input=step_input,
                                 spec={TEST_OUTPUT_PATH_KEY: f'{test_result_path}'})



""" This module is called on to create a helm chart for your project and install it during the `mpyl.steps.deploy`
step.
"""

import shutil
from logging import Logger
from pathlib import Path

from .resources import to_yaml, CustomResourceDefinition
from ...models import RunProperties, Output, Input
from ....utilities.subprocess import custom_check_output


def to_chart_metadata(chart_name: str, run_properties: RunProperties):
    return f"""apiVersion: v3
name: {chart_name}
description: A helm chart used by the MPyL pipeline
type: application
version: 0.1.0
appVersion: "{run_properties.versioning.identifier}"
"""


def write_chart(chart: dict[str, CustomResourceDefinition], chart_path: Path, chart_metadata: str) -> None:
    shutil.rmtree(chart_path, ignore_errors=True)
    template_path = chart_path / Path("templates")
    template_path.mkdir(parents=True, exist_ok=True)

    with open(chart_path / Path("Chart.yaml"), mode='w+', encoding='utf-8') as file:
        file.write(chart_metadata)
    with open(chart_path / Path("values.yaml"), mode='w+', encoding='utf-8') as file:
        file.write("# This file is intentionally left empty. All values in /templates have been pre-interpolated")

    my_dictionary: dict[str, str] = dict(map(lambda item: (item[0], to_yaml(item[1])), chart.items()))

    for name, template in my_dictionary.items():
        with open(template_path / name, mode='w+', encoding='utf-8') as file:
            file.write(template)


def __remove_existing_chart(logger: Logger, chart_name: str, name_space: str, kube_context: str) -> Output:
    found_chart = custom_check_output(logger, f"helm list -f {chart_name} -n {name_space}", capture_stdout=True)
    if chart_name in found_chart.message:
        cmd = f"helm uninstall {chart_name} -n {name_space} --kube-context {kube_context}"
        return custom_check_output(Logger("helm"), cmd)
    return Output(success=True, message=f"No existing chart {chart_name} found to delete")


def install(logger: Logger, chart: dict[str, CustomResourceDefinition], step_input: Input, chart_name: str,
            name_space: str, kube_context: str, delete_existing: bool = False) -> Output:
    if delete_existing:
        removed = __remove_existing_chart(logger, chart_name, name_space, kube_context)
        if not removed.success:
            return removed

    chart_path = Path(step_input.project.target_path) / "chart"
    write_chart(chart, chart_path, to_chart_metadata(chart_name, step_input.run_properties))

    cmd = f"helm upgrade -i {chart_name} -n {name_space} --kube-context {kube_context} {chart_path}"
    if step_input.dry_run:
        cmd = f"helm upgrade -i {chart_name} -n namespace --kube-context {kube_context} {chart_path} --debug --dry-run"

    return custom_check_output(logger, cmd)



"""
Data classes for the composition of Custom Resource Definitions.
More info: https://kubernetes.io/docs/concepts/extend-kubernetes/api-extension/custom-resources/
"""

from dataclasses import dataclass
from typing import Dict, Optional

from kubernetes.client import V1Deployment, V1Container, V1DeploymentSpec, V1ObjectMeta, V1PodSpec, \
    V1RollingUpdateDeployment, V1LabelSelector, V1ContainerPort, V1EnvVar, V1Service, \
    V1ServiceSpec, V1ServicePort, V1ServiceAccount, V1LocalObjectReference, \
    V1EnvVarSource, V1SecretKeySelector, V1Probe, ApiClient, V1HTTPGetAction, V1ResourceRequirements, \
    V1PodTemplateSpec, V1DeploymentStrategy, V1Job, V1JobSpec, V1CronJob, V1CronJobSpec, V1JobTemplateSpec, V1ConfigMap
from ruamel.yaml import YAML

from . import get_namespace
from .resources import CustomResourceDefinition, to_dict  # pylint: disable = no-name-in-module
from .resources.sealed_secret import V1SealedSecret
from .resources.spark import to_spark_body, get_spark_config_map_data, V1SparkApplication
from .resources.traefik import V1AlphaIngressRoute, V1AlphaMiddleware, \
    HostWrapper  # pylint: disable = no-name-in-module
from ...models import Input, ArtifactType
from ....project import Project, KeyValueProperty, Probe, Deployment, TargetProperty, Resources, Target, Kubernetes, \
    Job, Traefik, Host
from ....utilities.ephemeral import get_env_variables

yaml = YAML()

# Determined (unscientifically) to be sensible factors.
# Based on actual CPU usage, pods rarely use more than 10% of the allocated CPU. 60% usage is healthy, so we
# scale down to 20% in order to keep some slack.
# Memory is cheaper, but poses a harder limit (OOM when exceeding limit), so we are more generous than with CPU.
CPU_REQUEST_SCALE_FACTOR = 0.2
MEM_REQUEST_SCALE_FACTOR = 0.5


def try_parse_target(value: object, target: Target):
    if isinstance(value, dict):
        maybe_value = TargetProperty.from_config(value).get_value(target)
        if maybe_value:
            return maybe_value

    return value


def with_target(dictionary: dict, target: Target) -> dict:
    def with_targets_parsed(obj):
        if isinstance(obj, dict):
            return type(obj)((k, try_parse_target(with_targets_parsed(v), target)) for k, v in obj.items())

        return obj

    return with_targets_parsed(dictionary)


@dataclass(frozen=True)
class ResourceDefaults:
    instances: TargetProperty[int]
    cpus: TargetProperty[float]
    mem: TargetProperty[int]

    @staticmethod
    def from_config(resources: dict):
        limit = resources['limit']
        return ResourceDefaults(instances=TargetProperty.from_config(resources['instances']),
                                cpus=TargetProperty.from_config(limit['cpus']),
                                mem=TargetProperty.from_config(limit['mem']))


@dataclass(frozen=True)
class DeploymentDefaults:
    resources_defaults: ResourceDefaults
    liveness_probe_defaults: dict
    startup_probe_defaults: dict
    job_defaults: dict
    treafik_defaults: dict
    white_lists: dict

    @staticmethod
    def from_config(config: dict):
        deployment_values = config.get('project', {}).get('deployment', {})
        if deployment_values is None:
            raise KeyError("Configuration should have project.deployment section")
        kubernetes = deployment_values.get('kubernetes', {})
        return DeploymentDefaults(
            resources_defaults=ResourceDefaults.from_config(kubernetes['resources']),
            liveness_probe_defaults=kubernetes['livenessProbe'],
            startup_probe_defaults=kubernetes['startupProbe'],
            job_defaults=kubernetes.get('job', {}),
            treafik_defaults=deployment_values.get('traefik', {}),
            white_lists=config.get('whiteLists', {})
        )


class ChartBuilder:
    step_input: Input
    project: Project
    mappings: dict[int, int]
    env: list[KeyValueProperty]
    sealed_secrets: list[KeyValueProperty]
    deployment: Deployment
    target: Target
    release_name: str
    config_defaults: DeploymentDefaults

    def __init__(self, step_input: Input):
        self.step_input = step_input
        project = self.step_input.project
        self.project = project
        if project.deployment is None:
            raise AttributeError("deployment field should be set")

        self.config_defaults = DeploymentDefaults.from_config(step_input.run_properties.config)

        self.deployment = project.deployment
        properties = self.deployment.properties
        self.env = properties.env if properties and properties.env else []
        self.sealed_secrets = properties.sealed_secret if properties and properties.sealed_secret else []
        self.mappings = self.project.kubernetes.port_mappings
        self.target = step_input.run_properties.target
        self.release_name = self.project.name.lower()

    def _to_labels(self) -> Dict:
        run_properties = self.step_input.run_properties
        app_labels = {'name': self.release_name, 'app.kubernetes.io/version': run_properties.versioning.identifier,
                      'app.kubernetes.io/managed-by': 'Helm', 'app.kubernetes.io/name': self.release_name,
                      'app.kubernetes.io/instance': self.release_name}

        if len(self.project.maintainer) > 0:
            app_labels['maintainers'] = ".".join(self.project.maintainer).replace(' ', '_')
            app_labels["maintainer"] = self.project.maintainer[0].replace(' ', '_')

        app_labels['version'] = run_properties.versioning.identifier

        if run_properties.versioning.revision:
            app_labels['revision'] = run_properties.versioning.revision

        return app_labels

    def _to_annotations(self) -> Dict:
        return {'description': self.project.description}

    def _to_object_meta(self, name: Optional[str] = None):
        return V1ObjectMeta(name=name if name else self.release_name, labels=self._to_labels())

    def _to_selector(self):
        return V1LabelSelector(match_labels={"app.kubernetes.io/instance": self.release_name,
                                             "app.kubernetes.io/name": self.release_name})

    @staticmethod
    def _to_k8s_model(values: dict, model_type):
        return ApiClient()._ApiClient__deserialize(values, model_type)  # pylint: disable=protected-access

    @staticmethod
    def _to_probe(probe: Probe, defaults: dict, target: Target) -> V1Probe:
        values = defaults.copy()
        values.update(probe.values)
        v1_probe: V1Probe = ChartBuilder._to_k8s_model(values, V1Probe)
        path = probe.path.get_value(target)
        v1_probe.http_get = V1HTTPGetAction(path='/health' if path is None else path, port='port-0')
        return v1_probe

    def to_service(self) -> V1Service:
        service_ports = list(map(lambda key: V1ServicePort(port=key, target_port=self.mappings[key], protocol="TCP",
                                                           name=f"{key}-webservice-port"), self.mappings.keys()))

        return V1Service(api_version='v1', kind='Service', metadata=self._to_object_meta(),
                         spec=V1ServiceSpec(type="ClusterIP", ports=service_ports,
                                            selector=self._to_selector().match_labels))

    def to_job(self) -> V1Job:
        job_container = V1Container(
            name=self.release_name, image=self._get_image(), env=self._get_env_vars(), image_pull_policy="Always",
            resources=self._get_resources()
        )
        pod_template = V1PodTemplateSpec(
            metadata=self._to_object_meta(),
            spec=V1PodSpec(containers=[job_container], service_account=self.release_name,
                           service_account_name=self.release_name, restart_policy="Never")
        )

        defaults = with_target(self.config_defaults.job_defaults, self.target)
        specified = defaults | with_target(self.project.job.job, self.target)

        template_dict = to_dict(pod_template)
        specified['template'] = template_dict
        spec: V1JobSpec = ChartBuilder._to_k8s_model(specified, V1JobSpec)

        return V1Job(api_version='batch/v1', kind='Job', metadata=self._to_object_meta(), spec=spec)

    def to_cron_job(self) -> V1CronJob:
        values = self.project.job.cron
        job_template = V1JobTemplateSpec(spec=self.to_job().spec)
        template_dict = to_dict(job_template)
        values['jobTemplate'] = template_dict
        v1_cron_job_spec: V1CronJobSpec = ChartBuilder._to_k8s_model(values, V1CronJobSpec)
        return V1CronJob(api_version='batch/v1', kind='CronJob', metadata=self._to_object_meta(), spec=v1_cron_job_spec)

    def to_spark_application(self) -> V1SparkApplication:
        return V1SparkApplication(
            schedule=self._get_job().cron['schedule'],
            body=to_spark_body(
                project_name=self.release_name,
                env_vars=get_env_variables(self.project, self.target),
                spark=self._get_job().spark
            ),
        )

    def to_spark_config_map(self) -> V1ConfigMap:
        return V1ConfigMap(
            api_version='v1',
            kind='ConfigMap',
            data=get_spark_config_map_data(),
            metadata=self._to_object_meta()
        )

    def __find_default_port(self) -> int:
        found = next(iter(self.mappings.keys()))
        if found:
            return int(found)
        raise KeyError("No default port found. Did you define a port mapping?")

    def create_host_wrappers(self) -> list[HostWrapper]:
        default_hosts: list[Host] = Traefik.from_config(self.config_defaults.treafik_defaults).hosts

        hosts: list[Host] = self.deployment.traefik.hosts if self.deployment.traefik else []

        first_host = next(iter(hosts), None)
        service_port = first_host.service_port if first_host and first_host.service_port else self.__find_default_port()

        def to_white_list(configured: Optional[TargetProperty[list[str]]]) -> list[str]:
            white_lists = configured.get_value(self.target) if configured else self.config_defaults.white_lists[
                'default']
            addresses = []
            add_dict = dict(
                (address['name'], address['values']) for address in self.config_defaults.white_lists['addresses'])
            for item in white_lists:
                addresses.extend(add_dict[item])

            return addresses

        return [HostWrapper(host=host, name=self.release_name, index=idx, service_port=service_port,
                            white_lists=to_white_list(host.whitelists)) for
                idx, host in enumerate(hosts if hosts else default_hosts)]

    def to_ingress_routes(self) -> V1AlphaIngressRoute:
        hosts = self.create_host_wrappers()

        return V1AlphaIngressRoute(metadata=self._to_object_meta(), hosts=hosts, target=self.target,
                                   pr_number=self.step_input.run_properties.versioning.pr_number)

    def to_middlewares(self) -> dict[str, V1AlphaMiddleware]:
        hosts = self.create_host_wrappers()
        return dict(map(lambda host:
                        (host.full_name,
                         V1AlphaMiddleware(metadata=self._to_object_meta(name=host.full_name),
                                           source_ranges=host.white_lists)),
                        hosts))

    def to_service_account(self) -> V1ServiceAccount:
        return V1ServiceAccount(api_version="v1", kind="ServiceAccount", metadata=self._to_object_meta(),
                                image_pull_secrets=[V1LocalObjectReference("bigdataregistry")])

    def to_sealed_secrets(self) -> V1SealedSecret:
        secrets: dict[str, str] = {}
        for secret in self.sealed_secrets:
            secrets[secret.key] = secret.get_value(self.target)

        return V1SealedSecret(name=self.release_name, secrets=secrets)

    @staticmethod
    def _to_resources(resources: Resources, defaults: ResourceDefaults, target: Target):
        cpus = resources.cpus if resources and resources.cpus else defaults.cpus
        cpus_limit = cpus.get_value(target=target) * 1000.0
        cpus_request = cpus_limit * CPU_REQUEST_SCALE_FACTOR

        mem = resources.mem if resources and resources.mem else defaults.mem
        mem_limit = mem.get_value(target=target)
        mem_request = mem_limit * MEM_REQUEST_SCALE_FACTOR
        return V1ResourceRequirements(limits={'cpu': f'{int(cpus_limit)}m', 'memory': f'{int(mem_limit)}Mi'},
                                      requests={'cpu': f'{int(cpus_request)}m', 'memory': f'{int(mem_request)}Mi'})

    def _get_image(self):
        docker_image = self.step_input.required_artifact
        if not docker_image or docker_image.artifact_type != ArtifactType.DOCKER_IMAGE:
            raise ValueError(
                f'Required artifact of type {ArtifactType.DOCKER_IMAGE.name} must be defined')  # pylint: disable=E1101
        return docker_image.spec['image']

    def _get_resources(self):
        resources = self.project.kubernetes.resources
        defaults = self.config_defaults.resources_defaults
        return ChartBuilder._to_resources(resources, defaults, self.target)

    def _get_kubernetes(self) -> Kubernetes:
        kubernetes = self.deployment.kubernetes
        if kubernetes is None:
            raise AttributeError("deployment.kubernetes field should be set")
        return kubernetes

    def _get_job(self) -> Job:
        job = self._get_kubernetes().job
        if job is None:
            raise AttributeError("deployment.kubernetes.job field should be set")
        return job

    def _get_env_vars(self):

        def _interpolate_namespace(value: Optional[str]) -> Optional[str]:
            if value and '{namespace}' in value:
                namespace = get_namespace(self.step_input.run_properties, self.step_input.project)
                return value.replace('{namespace}', namespace)

            return value

        env_vars = list(
            filter(lambda v: v.value,
                   map(lambda e: V1EnvVar(name=e.key, value=_interpolate_namespace(e.get_value(self.target))),
                       self.env)))

        sealed_for_target = list(
            filter(lambda v: v.get_value(self.target) is not None, self.sealed_secrets))
        sealed_secrets = list(map(lambda e: V1EnvVar(name=e.key, value_from=V1EnvVarSource(
            secret_key_ref=V1SecretKeySelector(key=e.key, name=self.release_name, optional=False))),
                                  sealed_for_target))
        return env_vars + sealed_secrets

    @property
    def is_cron_job(self) -> bool:
        return len(self._get_job().cron.keys()) > 0

    def to_deployment(self) -> V1Deployment:

        ports = [
            V1ContainerPort(container_port=key, host_port=self.mappings[key], protocol="TCP", name=f'port-{idx}')
            for idx, key in enumerate(self.mappings.keys())
        ]

        project = self.project
        resources = project.resources
        kubernetes = project.kubernetes
        defaults = self.config_defaults.resources_defaults

        container = V1Container(
            name=self.release_name,
            image=self._get_image(),
            env=self._get_env_vars(),
            ports=ports,
            image_pull_policy="Always",
            resources=ChartBuilder._to_resources(resources, defaults, self.target),
            liveness_probe=ChartBuilder._to_probe(
                kubernetes.liveness_probe,
                self.config_defaults.liveness_probe_defaults,
                self.target
            ) if kubernetes.liveness_probe else None,
            startup_probe=ChartBuilder._to_probe(
                kubernetes.startup_probe,
                self.config_defaults.startup_probe_defaults,
                self.target)
            if kubernetes.startup_probe else None
        )

        instances = resources.instances if resources.instances else defaults.instances

        return V1Deployment(
            api_version="apps/v1",
            kind="Deployment",
            metadata=V1ObjectMeta(annotations=self._to_annotations(), name=self.release_name,
                                  labels=self._to_labels()),
            spec=V1DeploymentSpec(
                replicas=instances.get_value(target=self.target),
                template=V1PodTemplateSpec(
                    metadata=self._to_object_meta(),
                    spec=V1PodSpec(containers=[container], service_account=self.release_name,
                                   service_account_name=self.release_name),
                ),
                strategy=V1DeploymentStrategy(
                    rolling_update=V1RollingUpdateDeployment(max_surge="25%", max_unavailable="25%"),
                    type="RollingUpdate"),
                selector=self._to_selector(),
            ),
        )

    def to_common_chart(self) -> dict[str, CustomResourceDefinition]:
        chart = {'service-account': self.to_service_account()}

        if self.sealed_secrets:
            chart['sealed-secrets'] = self.to_sealed_secrets()

        return chart


def to_service_chart(builder: ChartBuilder) -> dict[str, CustomResourceDefinition]:
    return builder.to_common_chart() | {
        'deployment': builder.to_deployment(),
        'service': builder.to_service(),
        'ingress-https-route': builder.to_ingress_routes()
    } | builder.to_middlewares()


def to_job_chart(builder: ChartBuilder) -> dict[str, CustomResourceDefinition]:
    return builder.to_common_chart() | {'job': builder.to_job()}


def to_cron_job_chart(builder: ChartBuilder) -> dict[str, CustomResourceDefinition]:
    return builder.to_common_chart() | {'cronjob': builder.to_cron_job()}


def to_spark_job_chart(builder: ChartBuilder) -> dict[str, CustomResourceDefinition]:
    return builder.to_common_chart() | {
        'spark': builder.to_spark_application(),
        'config-map': builder.to_spark_config_map()
    }



""" A dummy deploy step, which produces `mpyl.steps.models.ArtifactType.NONE`."""

from logging import Logger

from .. import Step, Meta
from ..models import Input, Output, ArtifactType
from ...project import Stage


class DeployEcho(Step):

    def __init__(self, logger: Logger) -> None:
        super().__init__(logger, Meta(
            name='Echo Deploy',
            description='Dummy deploy step to test the framework',
            version='0.0.1',
            stage=Stage.DEPLOY
        ), ArtifactType.NONE, ArtifactType.DOCKER_IMAGE)

    def execute(self, step_input: Input) -> Output:
        artifact = step_input.required_artifact
        if not artifact:
            return Output(success=False, message=f"Step requires artifact of type {self.required_artifact}")

        self._logger.info(f"Deploying project {step_input.project.name} with artifact: {artifact.spec}")
        return Output(success=True, message=f"Deployed project {step_input.project.name}", produced_artifact=None)



"""Wrapper around `junitparser`"""
import os
from dataclasses import dataclass
from pathlib import Path

from junitparser import JUnitXml, TestSuite

from ...steps.models import Artifact, ArtifactType

TEST_OUTPUT_PATH_KEY = 'test_output_path'


@dataclass(frozen=True)
class TestRunSummary:
    tests: int
    failures: int
    errors: int
    skipped: int

    @property
    def is_success(self):
        return self.errors == 0 and self.failures == 0


def to_test_suites(artifact: Artifact) -> list[TestSuite]:
    if artifact.artifact_type != ArtifactType.JUNIT_TESTS:
        raise ValueError(f'Artifact {artifact} should be of type {ArtifactType.JUNIT_TESTS}')
    junit_result_path = artifact.spec[TEST_OUTPUT_PATH_KEY]

    xml = JUnitXml()
    for file_name in [fn for fn in os.listdir(junit_result_path) if fn.endswith('.xml')]:
        xml += JUnitXml.fromfile(Path(junit_result_path, file_name).as_posix())

    suites = [TestSuite.fromelem(s) for s in xml]
    return sorted(suites, key=lambda s: s.time)


def sum_suites(suites: list[TestSuite]) -> TestRunSummary:
    return TestRunSummary(tests=sum(s.tests for s in suites), failures=sum(s.failures for s in suites),
                          errors=sum(s.errors for s in suites), skipped=sum(s.skipped for s in suites))



"""Commands related to build"""
import asyncio
import shutil
from pathlib import Path

import click
from click import ParamType, BadParameter
from click.shell_completion import CompletionItem
from rich.console import Console
from rich.markdown import Markdown

from . import CliContext, CONFIG_PATH_HELP, check_updates, get_meta_version
from . import create_console_logger
from .commands.build.jenkins import JenkinsRunParameters, run_jenkins
from .commands.build.mpyl import MpylRunParameters, run_mpyl, MpylCliParameters, MpylRunConfig, find_build_set
from ..project import load_project
from ..reporting.formatting.markdown import run_result_to_markdown
from ..steps.models import RunProperties
from ..steps.run import RunResult
from ..utilities.pyaml_env import parse_config
from ..utilities.repo import Repository, RepoConfig


async def warn_if_update(console: Console):
    version = get_meta_version()
    update = await check_updates(meta=version)
    if update:
        console.print(Markdown(f"⚠️  **You can upgrade from {version} to {update} :** `pip install -U mpyl=={update}`"))


@click.group('build')
@click.option('--config', '-c', required=True, type=click.Path(exists=True), help=CONFIG_PATH_HELP,
              envvar="MPYL_CONFIG_PATH", default='mpyl_config.yml')
@click.option('--verbose', '-v', is_flag=True, default=False)
@click.pass_context
def build(ctx, config, verbose):
    """Pipeline build commands"""
    console = create_console_logger(local=False, verbose=verbose)
    parsed_config = parse_config(config)
    repo = ctx.with_resource(Repository(config=RepoConfig(parsed_config)))
    ctx.obj = CliContext(parsed_config, repo, console, verbose)


@build.command(help='Run an MPyL build')
@click.option('--properties', '-p', required=False, type=click.Path(exists=False), help='Path to run properties',
              envvar="MPYL_RUN_PROPERTIES_PATH", default='run_properties.yml')
@click.option('--ci', is_flag=True,
              help='Run as CI build instead of local. Ignores unversioned changes.')
@click.option('--all', 'all_', is_flag=True, help='Build all projects, regardless of changes on branch')
@click.pass_obj
def run(obj: CliContext, properties, ci, all_):  # pylint: disable=invalid-name
    asyncio.run(warn_if_update(obj.console))
    run_properties = RunProperties.from_configuration(parse_config(properties), obj.config) if ci \
        else RunProperties.for_local_run(obj.config, obj.repo.get_sha, obj.repo.get_branch)

    parameters = MpylCliParameters(local=not ci, pull_main=all_, all=all_, verbose=obj.verbose)
    obj.console.log(parameters)
    run_parameters = MpylRunParameters(
        run_config=MpylRunConfig(config=obj.config, run_properties=run_properties),
        parameters=parameters
    )
    run_mpyl(run_parameters, None)


@build.command(help="The status of the current local branch from MPyL's perspective")
@click.pass_obj
def status(obj: CliContext):
    asyncio.run(warn_if_update(obj.console))
    branch = obj.repo.get_branch
    if obj.repo.main_branch == obj.repo.get_branch:
        obj.console.log(f'On main branch ({branch}), cannot determine build status')
        return

    changes_in_branch = obj.repo.changes_in_branch_including_local()
    build_set = find_build_set(obj.repo, changes_in_branch, False)
    run_properties = RunProperties.for_local_run(obj.config, obj.repo.get_sha, obj.repo.get_branch)
    result = RunResult(run_properties=run_properties, run_plan=build_set)
    version = run_properties.versioning
    header: str = f"**Revision:** `{version.branch}` at `{version.revision}`  \n"
    if result.run_plan:
        obj.console.print(Markdown(markup=header + "**Execution plan:**  \n" + run_result_to_markdown(result)))
    else:
        obj.console.print("No changes detected, nothing to do.")


class Pipeline(ParamType):
    name = 'pipeline'

    def shell_complete(self, ctx: click.Context, param, incomplete: str):
        if not ctx.parent or ctx.parent.params['config'] is None or not Path(ctx.parent.params['config']).exists():
            raise BadParameter('Either --config parameter must or MPYL_CONFIG_PATH env var must be set', ctx=ctx,
                               param=param)

        config: dict = parse_config(ctx.parent.params['config'])
        parsed_config: dict[str, str] = config['jenkins']['pipelines']

        return [
            CompletionItem(value=pl[0], help=pl[1]) for pl in parsed_config.items() if
            incomplete in pl[0]
        ]


@build.command(help='Run a multi branch pipeline build on Jenkins')
@click.option(
    '--user', '-u',
    help='Authentication API user. Can be set via env var JENKINS_USER',
    envvar="JENKINS_USER",
    type=click.STRING,
    required=True
)
@click.option(
    '--password', '-p',
    help='Authentication API password. Can be set via env var JENKINS_PASSWORD',
    envvar="JENKINS_PASSWORD",
    type=click.STRING,
    required=True
)
@click.option(
    '--pipeline', '-pl',
    help='The pipeline to run. Must be one of the pipelines listed in `jenkins.pipelines`. '
         'Default value is `jenkins.defaultPipeline`',
    type=Pipeline(),
    required=False
)
@click.pass_context
def jenkins(ctx, user, password, pipeline):
    asyncio.run(warn_if_update(ctx.obj.console))
    selected_pipeline = pipeline if pipeline else ctx.obj.config['jenkins']['defaultPipeline']
    run_argument = JenkinsRunParameters(user, password, ctx.obj.config, selected_pipeline, ctx.obj.verbose)
    run_jenkins(run_argument)


@build.command(help='Clean MPyL metadata in `.mpl` folders')
@click.option('--filter', '-f', 'filter_', required=False, type=click.STRING, help='Filter based on filepath ')
@click.pass_obj
def clean(obj: CliContext, filter_):
    found_projects = obj.repo.find_projects(filter_ if filter_ else '')
    for project in found_projects:
        target_path = Path(load_project(obj.repo.root_dir(), Path(project), strict=False).target_path)
        if target_path.exists():
            shutil.rmtree(target_path)
            obj.console.print(f"🧹 Cleaned up {target_path}")


if __name__ == '__main__':
    build()  # pylint: disable=no-value-for-parameter



"""SBT config"""
import ast
from dataclasses import dataclass
from typing import Dict


@dataclass(frozen=True)
class SbtConfig:
    java_opts: str
    sbt_opts: str
    sbt_command: str
    sbt_client_command: str
    test_with_coverage: bool
    verbose: bool
    build_with_client: bool
    test_with_client: bool

    @staticmethod
    def from_config(config: Dict):
        sbt_config = config.get('sbt', None)
        if not sbt_config:
            raise KeyError(f"'sbt' could not be loaded from {config}")
        return SbtConfig(
            sbt_command=sbt_config['command'],
            sbt_client_command=sbt_config['clientCommand'],
            java_opts=sbt_config['javaOpts'],
            sbt_opts=sbt_config['sbtOpts'],
            test_with_coverage=(str(sbt_config['testWithCoverage']).lower() == 'true'),
            verbose=(str(sbt_config['verbose']).lower() == 'true'),
            build_with_client=(str(sbt_config.get('clientMode', {}).get('build')).lower() == 'true'),
            test_with_client=(str(sbt_config.get('clientMode', {}).get('test')).lower() == 'true')
        )

    def to_command(self, client_mode: bool, sbt_commands: list[str]):
        cmd = [self.sbt_client_command if client_mode else self.sbt_command]
        if self.verbose:
            cmd.append('-v')
        cmd.extend([f'-J{opt}' for opt in self.java_opts.split(' ')])
        cmd.extend([f'-D{opt}' for opt in self.sbt_opts.split(' ')])

        joined_commands = "; ".join(sbt_commands)
        cmd.append(joined_commands)
        return cmd



"""A step to compile and run tests for an SBT project"""
from logging import Logger
from typing import Callable

from .after_test import IntegrationTestAfter
from .before_test import IntegrationTestBefore
from .. import Input, Output, Step
from ..models import Artifact, input_to_artifact
from ...project import Stage, Project
from ...steps import Meta, ArtifactType
from ...utilities.junit import TEST_OUTPUT_PATH_KEY, to_test_suites, sum_suites
from ...utilities.sbt import SbtConfig
from ...utilities.subprocess import custom_check_output


class TestSbt(Step):
    def __init__(self, logger: Logger) -> None:
        meta = Meta(name='Sbt Test', description='Run sbt tests', version='0.0.1', stage=Stage.TEST)
        super().__init__(
            logger=logger,
            meta=meta,
            produced_artifact=ArtifactType.JUNIT_TESTS,
            required_artifact=ArtifactType.NONE,
            before=IntegrationTestBefore(logger),
            after=IntegrationTestAfter(logger)
        )

    def _test_with_coverage(self, step_input: Input, sbt_config: SbtConfig) -> Output:
        command_compile = self._construct_sbt_command(step_input, sbt_config,
                                                      self._construct_sbt_command_compile_with_coverage)
        compile_outcome = custom_check_output(self._logger, command_compile)
        project_name = step_input.project.name
        if not compile_outcome.success:
            return Output(success=False, message=f"Tests failed to compile for {project_name}",
                          produced_artifact=None)

        command_test = self._construct_sbt_command(step_input, sbt_config,
                                                   self._construct_sbt_command_test_with_coverage)
        test_outcome = custom_check_output(self._logger, command_test)
        artifact = self._extract_test_report(step_input.project, step_input)
        if not test_outcome.success:
            return Output(success=False,
                          message=f"Tests failed to run for {project_name}. No test results have been recorded.",
                          produced_artifact=artifact)
        return Output(success=True, message="Success", produced_artifact=artifact)

    def _test_without_coverage(self, step_input: Input, sbt_config: SbtConfig) -> Output:
        command_test_without_coverage = self._construct_sbt_command(step_input, sbt_config,
                                                                    self._construct_sbt_command_test_without_coverage)
        run_outcome = custom_check_output(self._logger, command_test_without_coverage)
        artifact = self._extract_test_report(step_input.project, step_input)
        if not run_outcome.success:
            return Output(success=False, message=f"Tests without coverage failed to run for {step_input.project.name}",
                          produced_artifact=artifact)
        return Output(success=True, message="Success", produced_artifact=artifact)

    def execute(self, step_input: Input) -> Output:
        project = step_input.project
        sbt_config = SbtConfig.from_config(config=step_input.run_properties.config)
        self._logger.debug(f'Config {sbt_config}')

        test_result = self._test_with_coverage(step_input, sbt_config) if sbt_config.test_with_coverage \
            else self._test_without_coverage(step_input, sbt_config)

        if test_result.produced_artifact:
            suite = to_test_suites(test_result.produced_artifact)
            summary = sum_suites(suite)
            return Output(success=summary.is_success,
                          message=f"Tests results produced for {project.name} ({summary})",
                          produced_artifact=test_result.produced_artifact)

        return test_result

    @staticmethod
    def _construct_sbt_command_compile_with_coverage(step_input: Input) -> list[str]:
        return [
            f'project {step_input.project.name}',
            'coverageOn',
            'test:compile'
        ]

    @staticmethod
    def _construct_sbt_command_test_with_coverage(step_input: Input):
        return [
            f'project {step_input.project.name}',
            'test',
            'coverageOff'
        ]

    @staticmethod
    def _construct_sbt_command_test_without_coverage(step_input: Input):
        return [f'{step_input.project.name}/test']

    @staticmethod
    def _construct_sbt_command(step_input: Input, config: SbtConfig, commands_fn: Callable[[Input], list[str]]):
        return config.to_command(config.test_with_client, commands_fn(step_input))

    @staticmethod
    def _extract_test_report(project: Project, step_input: Input) -> Artifact:
        return input_to_artifact(artifact_type=ArtifactType.JUNIT_TESTS, step_input=step_input,
                                 spec={TEST_OUTPUT_PATH_KEY: f'{project.test_report_path}'})



""" Step that builds a docker image from a project in a multi-project sbt. Assumes that the
[sbt-docker](https://github.com/marcuslonnberg/sbt-docker) plugin is present.
"""

from logging import Logger
from typing import Optional

from .docker_after_build import AfterBuildDocker
from .. import Step, Meta
from ...project import Stage, Target
from ...steps.models import ArtifactType, Input, Output, input_to_artifact
from ...utilities.docker import docker_image_tag
from ...utilities.sbt import SbtConfig
from ...utilities.subprocess import custom_check_output


class BuildSbt(Step):
    def __init__(self, logger: Logger) -> None:
        self.logger = logger
        super().__init__(
            logger=logger,
            meta=Meta(
                name='Sbt Build',
                description='Build sbt project',
                version='0.0.1',
                stage=Stage.BUILD
            ),
            produced_artifact=ArtifactType.DOCKER_IMAGE,
            required_artifact=ArtifactType.NONE,
            after=AfterBuildDocker(logger=logger)
        )

    def execute(self, step_input: Input) -> Output:
        image_name = docker_image_tag(step_input)
        command = self._construct_sbt_command(step_input, image_name)

        output = custom_check_output(self.logger, command=command)
        artifact = input_to_artifact(ArtifactType.DOCKER_IMAGE, step_input, {'image': image_name})
        if output.success:
            return Output(success=True, message=f"Built {step_input.project.name}", produced_artifact=artifact)

        return Output(success=False, message=f"Failed to build sbt project for {step_input.project.name}",
                      produced_artifact=None)

    @staticmethod
    def _construct_sbt_command(step_input: Input, image_name: str):
        check_fmt: Optional[str] = \
            'scalafmtCheckAll' if step_input.run_properties.target == Target.PULL_REQUEST else None
        commands: list[str] = [
            command for command in [
                f'project {step_input.project.name}',
                f'set docker / imageNames := Seq(ImageName("{image_name}"))',
                check_fmt,
                'docker'
            ] if command is not None
        ]
        config = SbtConfig.from_config(config=step_input.run_properties.config)
        command = config.to_command(config.build_with_client, commands)
        return command



"""Commands related to projects and how they relate"""
from dataclasses import dataclass
from pathlib import Path

import click
import jsonschema
from click import ParamType, BadParameter
from click.shell_completion import CompletionItem
from rich.markdown import Markdown

from . import CliContext, CONFIG_PATH_HELP, create_console_logger
from .commands.projects.formatting import print_project
from ..project import validate_project, load_project, Project
from ..utilities.pyaml_env import parse_config
from ..utilities.repo import Repository, RepoConfig


@dataclass
class ProjectsContext:
    cli: CliContext
    filter: str


@click.group('projects')
@click.option('--config', '-c', required=True, type=click.Path(exists=True), help=CONFIG_PATH_HELP,
              envvar="MPYL_CONFIG_PATH", default='mpyl_config.yml')
@click.option('--verbose', '-v', is_flag=True, default=False)
@click.option('--filter', '-f', 'filter_', required=False, type=click.STRING, help='Filter based on filepath ')
@click.pass_context
def projects(ctx, config, verbose, filter_):
    """Commands related to projects"""
    console = create_console_logger(local=False, verbose=verbose)
    parsed_config = parse_config(config)
    ctx.obj = ProjectsContext(
        cli=CliContext(config=parsed_config, repo=ctx.with_resource(Repository(config=RepoConfig(parsed_config))),
                       console=console, verbose=verbose), filter=filter_ if filter_ else '')


@projects.command(name='list', help='List found projects')
@click.pass_obj
def list_projects(obj: ProjectsContext):
    found_projects = obj.cli.repo.find_projects(obj.filter)

    for proj in found_projects:
        name = load_project(obj.cli.repo.root_dir(), Path(proj), False).name
        obj.cli.console.print(Markdown(f"{proj} `{name}`"))


class ProjectPath(ParamType):
    name = 'project_path'

    def shell_complete(self, ctx: click.Context, param, incomplete: str):
        if not ctx.parent or ctx.parent.params['config'] is None or not Path(ctx.parent.params['config']).exists():
            raise BadParameter('Either --config parameter must or MPYL_CONFIG_PATH env var must be set', ctx=ctx,
                               param=param)

        parsed_config = parse_config(ctx.parent.params['config'])
        repo = ctx.with_resource(Repository(config=RepoConfig(parsed_config)))
        found_projects = repo.find_projects(incomplete)
        return [
            CompletionItem(value=proj.replace(f'/{Project.project_yaml_path()}', ''))
            for proj in found_projects
        ]


@projects.command(name='show', help='Show details of a project')
@click.argument('name', required=True, type=ProjectPath())
@click.pass_obj
def show_project(obj, name):
    print_project(obj.cli.repo, obj.cli.console, f'{name}/{Project.project_yaml_path()}')


@projects.command(help='Validate the yaml of found projects against their schema')
@click.pass_obj
def lint(obj: ProjectsContext):
    for project in obj.cli.repo.find_projects(obj.filter):
        try:
            project_path = Path(obj.cli.repo.root_dir()) / Path(project)
            with open(project_path, encoding='utf-8') as file:
                validate_project(file)
        except jsonschema.exceptions.ValidationError as exc:
            obj.cli.console.print(f'❌ {project}: {exc.message}')
        else:
            obj.cli.console.print(f'✅ {project}')


if __name__ == '__main__':
    projects()  # pylint: disable=no-value-for-parameter



""" Model representation of run-specific configuration. """

import pkgutil
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Optional, Dict

from ruamel.yaml import YAML, yaml_object  # type: ignore

from ..project import Project, Stage, Target
from ..validation import validate

yaml = YAML()


@dataclass(frozen=True)
class VersioningProperties:
    revision: str
    branch: Optional[str]
    pr_number: Optional[int]
    tag: Optional[str]

    def __post_init__(self):
        if not self.pr_number and not self.tag:
            raise ValueError('Either pr_number or tag need to be set')

    @property
    def identifier(self):
        return f'pr-{self.pr_number}' if self.pr_number else self.tag


@dataclass(frozen=True)
class RunContext:
    build_id: str
    """Uniquely identifies the run. Typically a monotonically increasing number"""
    run_url: str
    """Link back to the run executor"""
    change_url: str
    """Link to changes"""
    tests_url: str
    """Link to test results"""
    user: str
    """Name of of the user that triggered the run"""
    user_email: Optional[str]
    """Email of of the user that triggered the run"""

    @staticmethod
    def from_configuration(run_details: Dict):
        return RunContext(build_id=run_details['id'], run_url=run_details['run_url'],
                          change_url=run_details['change_url'], tests_url=run_details['tests_url'],
                          user=run_details['user'], user_email=run_details['user_email'])


@yaml_object(yaml)
@dataclass(frozen=True)
class RunProperties:
    """ Contains information that is specific to a particular run of the pipeline"""
    details: RunContext
    """Run specific details"""
    target: Target
    """The deploy target"""
    versioning: VersioningProperties
    config: dict
    """Globally specified configuration, to be used by specific steps. Complies with the schema as
    specified in `mpyl_config.schema.yml`
     """

    @staticmethod
    def for_local_run(config: Dict, revision: str, branch: Optional[str]):
        return RunProperties(details=RunContext("", "", "", "", "", None), target=Target.PULL_REQUEST,
                             versioning=VersioningProperties(revision, branch, 123, None), config=config)

    @staticmethod
    def from_configuration(run_properties: Dict, config: Dict):
        build_dict = pkgutil.get_data(__name__, "../schema/run_properties.schema.yml")

        if build_dict:
            validate(run_properties, build_dict.decode('utf-8'))

        build = run_properties['build']

        versioning_config = build['versioning']

        pr_num: str = versioning_config.get('pr_number')
        tag: str = versioning_config.get('tag')

        versioning = VersioningProperties(revision=versioning_config['revision'],
                                          branch=versioning_config['branch'],
                                          pr_number=int(pr_num) if pr_num else None,
                                          tag=tag)

        return RunProperties(details=RunContext.from_configuration(build['run']),
                             target=Target(build['parameters']['deploy_target']), versioning=versioning, config=config)


@yaml_object(yaml)
@dataclass(frozen=True)
class ArtifactType(Enum):
    def __eq__(self, other):
        return self.value == other.value

    @classmethod
    def from_yaml(cls, _, node):
        return ArtifactType(int(node.value.split('-')[1]))

    @classmethod
    def to_yaml(cls, representer, node):
        return representer.represent_scalar('!ArtifactType',
                                            f'{node._name_}-{node._value_}')  # pylint: disable=protected-access

    DOCKER_IMAGE = 1
    """A docker image"""
    JUNIT_TESTS = 2
    """A test suite in junit compatible `.xml` format"""
    DEPLOYED_HELM_APP = 3
    """A helm chart deployed to kubernetes"""
    NONE = 4


@yaml_object(yaml)
@dataclass(frozen=True)
class Artifact:
    artifact_type: ArtifactType
    revision: str
    producing_step: str
    spec: dict


@yaml_object(yaml)
@dataclass(frozen=True)
class Input:
    project: Project
    run_properties: RunProperties
    """Run specific properties"""
    required_artifact: Optional[Artifact] = None
    dry_run: bool = False


@yaml_object(yaml)
@dataclass()
class Output:
    success: bool
    message: str
    produced_artifact: Optional[Artifact] = None

    @staticmethod
    def path(target_path: str, stage: Stage):
        return Path(target_path, f"{stage.name}.yml")

    def write(self, target_path: str, stage: Stage):
        Path(target_path).mkdir(parents=True, exist_ok=True)
        with Output.path(target_path, stage).open(mode='w+', encoding='utf-8') as file:
            yaml.dump(self, file)

    @staticmethod
    def try_read(target_path: str, stage: Stage):
        path = Output.path(target_path, stage)
        if path.exists():
            with open(path, encoding='utf-8') as file:
                return yaml.load(file)
        return None


def input_to_artifact(artifact_type: ArtifactType, step_input: Input, spec: dict):
    return Artifact(artifact_type=artifact_type, revision=step_input.run_properties.versioning.revision,
                    producing_step=step_input.project.name, spec=spec)



""" Utilities for creating rancher compatible helm charts. """

from dataclasses import dataclass

from ....steps import Input
from ....project import Target


@dataclass(frozen=True)
class ClusterConfig:
    project_id: str
    cluster_id: str
    cluster_env: str
    context: str

    @staticmethod
    def from_config(config: dict):
        return ClusterConfig(project_id=config['clusterId'], cluster_id=config['clusterId'],
                             cluster_env=config['clusterEnv'], context=config['context'])


def cluster_config(step_input: Input):
    target = step_input.run_properties.target
    cluster_configs = step_input.run_properties.config['kubernetes']['rancher']['cluster']

    if target in {Target.PULL_REQUEST, Target.PULL_REQUEST_BASE}:
        return ClusterConfig.from_config(cluster_configs['test'])
    if target == Target.ACCEPTANCE:
        return ClusterConfig.from_config(cluster_configs['acceptance'])
    if target == Target.PRODUCTION:
        return ClusterConfig.from_config(cluster_configs['production'])
    return None


def rancher_namespace_metadata(namespace: str, step_input: Input):
    rancher_config = cluster_config(step_input)

    return {
        'annotations': {
            'field.cattle.io/projectId': f'{rancher_config.cluster_id}:{rancher_config.project_id}',
            'lifecycle.cattle.io/create.namespace-auth': 'true'
        },
        'labels': {
            'field.cattle.io/projectId': rancher_config.project_id,
            'kubernetes.io/metadata.name': namespace
        },
        'name': namespace
    }



""" After test step. Stops any services running from docker-compose if necessary."""
import os
from logging import Logger

from python_on_whales import DockerClient

from .. import Step, Meta
from ..models import Input, Output, ArtifactType
from ...project import Stage
from ...utilities.docker import stream_docker_logging


class IntegrationTestAfter(Step):
    def __init__(self, logger: Logger) -> None:
        super().__init__(logger=logger,
                         meta=Meta(name='After Test', description='After test step', version='0.0.1', stage=Stage.TEST),
                         produced_artifact=ArtifactType.NONE, required_artifact=ArtifactType.NONE)

    def execute(self, step_input: Input) -> Output:
        compose_file = step_input.project.test_containers_path
        if not os.path.exists(compose_file):
            return Output(success=True, message='No containers to stop')

        self._logger.debug(f"Stopping containers in {compose_file}")
        docker_client = DockerClient(compose_files=[compose_file])
        logs = docker_client.compose.logs(stream=True)
        stream_docker_logging(logger=self._logger, generator=logs, task_name=f'Stop {compose_file}')

        docker_client.compose.down(remove_orphans=True)
        container_names = list(map(lambda l: l.name, docker_client.compose.ps(all=True)))

        return Output(success=True, message=f"Stopped {', '.join(container_names)}")



"""Ephemeral related utility methods"""

from ...project import Target, Project


def get_env_variables(project: Project, target: Target) -> dict[str, str]:
    if project.deployment is None:
        raise KeyError(f'No deployment information was found for project: {project.name}')
    if len(project.deployment.properties.env) == 0:
        raise KeyError(f'No properties.env is defined for project: {project.name}')

    env_variables: dict[str, str] = {
        env_variable.key: env_variable.get_value(target) for env_variable in
        project.deployment.properties.env
    }

    return env_variables



"""
Datamodel representation of project specific configuration as specified
in the `deployment/project.yml`. It defines how the source code to which it relates
"wants" to be built / tested / deployed.

<details>
  <summary>Schema definition</summary>
```yaml
.. include:: ./schema/project.schema.yml
```
</details>

.. include:: ../../README-dev.md
"""

import logging
import pkgutil
import time
import traceback
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Optional, TypeVar, Dict, Any, TextIO

import jsonschema
from mypy.checker import Generic
from ruamel.yaml import YAML

from .validation import validate

T = TypeVar('T')


def without_keys(dictionary: dict, keys: set[str]):
    return {k: dictionary[k] for k in dictionary.keys() - keys}


@dataclass(frozen=True)
class Target(Enum):
    def __eq__(self, other):
        return self.value == other.value

    PULL_REQUEST = 'PullRequest'
    PULL_REQUEST_BASE = 'PullRequestBase'
    ACCEPTANCE = 'Acceptance'
    PRODUCTION = 'Production'


@dataclass(frozen=True)
class Stage(Enum):
    def __eq__(self, other):
        return self.value == other.value

    BUILD = 'build'
    TEST = 'test'
    DEPLOY = 'deploy'
    POST_DEPLOY = 'postdeploy'


@dataclass(frozen=True)
class TargetProperty(Generic[T]):
    pr: Optional[T]  # pylint: disable=invalid-name
    test: Optional[T]
    acceptance: Optional[T]
    production: Optional[T]
    all: Optional[T]

    def get_value(self, target: Target):
        if self.all:
            return self.all
        if target == Target.PULL_REQUEST:
            return self.pr
        if target == Target.PULL_REQUEST_BASE:
            return self.test
        if target == Target.ACCEPTANCE:
            return self.acceptance
        if target == Target.PRODUCTION:
            return self.production
        return None

    @staticmethod
    def from_config(values: dict):
        if not values:
            return None
        return TargetProperty(pr=values.get('pr'), test=values.get('test'), acceptance=values.get('acceptance'),
                              production=values.get('production'), all=values.get('all'))


@dataclass(frozen=True)
class KeyValueProperty(TargetProperty[str]):
    key: str

    @staticmethod
    def from_config(values: dict):
        return KeyValueProperty(key=values['key'], pr=values.get('pr'), test=values.get('test'),
                                acceptance=values.get('acceptance'), production=values.get('production'),
                                all=values.get('all'))


@dataclass(frozen=True)
class StageSpecificProperty(Generic[T]):
    build: Optional[T]
    test: Optional[T]
    deploy: Optional[T]
    postdeploy: Optional[T]

    def for_stage(self, stage: Stage) -> Optional[T]:
        if stage == Stage.BUILD:
            return self.build
        if stage == Stage.TEST:
            return self.test
        if stage == Stage.DEPLOY:
            return self.deploy
        return self.postdeploy


@dataclass(frozen=True)
class Stages(StageSpecificProperty[str]):

    @staticmethod
    def from_config(values: dict):
        return Stages(build=values.get('build'), test=values.get('test'), deploy=values.get('deploy'),
                      postdeploy=values.get('postdeploy'))


@dataclass(frozen=True)
class Dependencies(StageSpecificProperty[set[str]]):

    def set_for_stage(self, stage: Stage) -> set[str]:
        deps_for_stage = self.for_stage(stage)
        return deps_for_stage if deps_for_stage else set()

    @staticmethod
    def from_config(values: dict):
        return Dependencies(build=set(values.get('build', [])), test=set(values.get('test', [])),
                            deploy=set(values.get('deploy', [])), postdeploy=set(values.get('postdeploy', [])))


@dataclass(frozen=True)
class Env:
    @staticmethod
    def from_config(values: list[dict]):
        return list(map(KeyValueProperty.from_config, values))


@dataclass(frozen=True)
class Properties:
    env: list[KeyValueProperty]
    sealed_secret: list[KeyValueProperty]

    @staticmethod
    def from_config(values: Dict[Any, Any]):
        return Properties(env=list(map(KeyValueProperty.from_config, values.get('env', []))),
                          sealed_secret=list(
                              map(KeyValueProperty.from_config, values.get('sealedSecret', []))))


@dataclass(frozen=True)
class Probe:
    path: TargetProperty[str]
    values: dict

    @staticmethod
    def from_config(values: dict):
        if not values:
            return None
        return Probe(path=TargetProperty.from_config(values['path']), values=values)


@dataclass(frozen=True)
class Metrics:
    path: str
    enabled: bool

    @staticmethod
    def from_config(values: dict):
        if not values:
            return None
        return Metrics(path=values.get('path', '/metrics'), enabled=values.get('enabled', False))


@dataclass(frozen=True)
class Resources:
    instances: Optional[TargetProperty[int]]
    cpus: Optional[TargetProperty[float]]
    mem: Optional[TargetProperty[int]]
    disk: Optional[TargetProperty[int]]

    @staticmethod
    def from_config(values: dict):
        limits = values.get('limit', {})
        return Resources(
            instances=TargetProperty.from_config(limits.get('instances', {})),
            cpus=TargetProperty.from_config(limits.get('cpus', {})),
            mem=TargetProperty.from_config(limits.get('mem', {})),
            disk=TargetProperty.from_config(limits.get('disk', {}))
        )


@dataclass(frozen=True)
class Job:
    cron: dict
    job: dict
    spark: dict

    @staticmethod
    def from_config(values: dict):
        if not values:
            return None
        return Job(cron=values.get('cron', {}), job=without_keys(values, {'cron'}), spark=values.get('spark', {}))


@dataclass(frozen=True)
class Kubernetes:
    port_mappings: dict[int, int]
    liveness_probe: Optional[Probe]
    startup_probe: Optional[Probe]
    metrics: Optional[Metrics]
    resources: Resources
    job: Optional[Job]

    @staticmethod
    def from_config(values: dict):
        return Kubernetes(
            port_mappings=values.get('portMappings', {}),
            liveness_probe=Probe.from_config(values.get('livenessProbe', {})),
            startup_probe=Probe.from_config(values.get('startupProbe', {})),
            metrics=Metrics.from_config(values.get('metrics', {})),
            resources=Resources.from_config(values.get('resources', {})),
            job=Job.from_config(values.get('job', {}))
        )


@dataclass(frozen=True)
class Host:
    host: TargetProperty[str]
    service_port: Optional[int]
    tls: TargetProperty[str]
    whitelists: TargetProperty[list[str]]

    @staticmethod
    def from_config(values: dict):
        return Host(
            host=TargetProperty.from_config(values.get('host', {})),
            service_port=values.get('servicePort'),
            tls=TargetProperty.from_config(values.get('tls', {})),
            whitelists=TargetProperty.from_config(values.get('whitelists', {}))
        )


@dataclass(frozen=True)
class Traefik:
    hosts: list[Host]

    @staticmethod
    def from_config(values: dict):
        hosts = values.get('hosts')
        return Traefik(hosts=(list(map(Host.from_config, hosts) if hosts else [])))


@dataclass(frozen=True)
class Deployment:
    namespace: Optional[str]
    properties: Properties
    kubernetes: Optional[Kubernetes]
    traefik: Optional[Traefik]

    @staticmethod
    def from_config(values: dict):
        props = values.get('properties')
        kubernetes = values.get('kubernetes')
        traefik = values.get('traefik')

        return Deployment(namespace=values.get('namespace'),
                          properties=Properties.from_config(props) if props else None,
                          kubernetes=Kubernetes.from_config(kubernetes) if kubernetes else None,
                          traefik=Traefik.from_config(traefik) if traefik else None)


@dataclass(frozen=True)
class Project:
    name: str
    description: str
    path: str
    stages: Stages
    maintainer: list[str]
    deployment: Optional[Deployment]
    dependencies: Optional[Dependencies]

    def __lt__(self, other):
        return self.path < other.path

    def __eq__(self, other):
        return self.path == other.path

    def __hash__(self):
        return hash(self.path)

    @property
    def kubernetes(self) -> Kubernetes:
        if self.deployment is None or self.deployment.kubernetes is None:
            raise KeyError(f"Project '{self.name}' does not have kubernetes configuration")
        return self.deployment.kubernetes

    @property
    def resources(self) -> Resources:
        return self.kubernetes.resources

    @property
    def job(self) -> Job:
        if self.kubernetes.job is None:
            raise KeyError(f"Project '{self.name}' does not have kubernetes.job configuration")
        return self.kubernetes.job

    @staticmethod
    def project_yaml_path() -> str:
        return 'deployment/project.yml'

    @property
    def root_path(self) -> str:
        return self.path.replace(Project.project_yaml_path(), '')

    @property
    def deployment_path(self) -> str:
        return str(Path(self.root_path, 'deployment'))

    @property
    def target_path(self) -> str:
        return str(Path(self.deployment_path, '.mpl'))

    @property
    def test_containers_path(self) -> str:
        return str(Path(self.deployment_path, 'docker-compose-test.yml'))

    @property
    def test_report_path(self) -> str:
        return str(Path(self.root_path, 'target/test-reports'))

    @staticmethod
    def from_config(values: dict, project_path: Path):
        deployment = values.get('deployment')
        dependencies = values.get('dependencies')
        return Project(name=values['name'], description=values['description'], path=str(project_path),
                       stages=Stages.from_config(values.get('stages', {})), maintainer=values.get('maintainer', []),
                       deployment=Deployment.from_config(deployment) if deployment else None,
                       dependencies=Dependencies.from_config(dependencies) if dependencies else None)


def validate_project(file: TextIO) -> dict:
    """
    :file the file to validate
    :return: the validated schema
    :raises `jsonschema.exceptions.ValidationError` when validation fails
    """
    yaml_values = YAML(typ='unsafe').load(file)
    template = pkgutil.get_data(__name__, "schema/project.schema.yml")
    if not template:
        raise ValueError('Schema project.schema.yml not found in package')
    validate(yaml_values, template.decode('utf-8'))

    return yaml_values


def load_project(root_dir: Path, project_path: Path, strict: bool = True) -> Project:
    """
    Load a `project.yml` to `Project` data class
    :param root_dir: root source directory
    :param project_path: relative path from `root_dir` to the `project.yml`
    :param strict: indicates whether the schema should be validated
    :return: `Project` data class
    """
    with open(root_dir / project_path, encoding='utf-8') as file:
        try:
            start = time.time()
            yaml_values = validate_project(file) if strict else YAML(typ='unsafe').load(file)
            project = Project.from_config(yaml_values, project_path)
            logging.debug(f'Loaded project {project.path} in {(time.time() - start) * 1000} ms')
            return project
        except jsonschema.exceptions.ValidationError as exc:
            logging.warning(f'{project_path} does not comply with schema: {exc.message}')
            raise
        except TypeError:
            traceback.print_exc()
            logging.warning('Type error', exc_info=True)
            raise
        except Exception:
            logging.warning(f'Failed to load {project_path}', exc_info=True)
            raise



"""
This module contains the Spark job CRD class.
"""
from typing import Optional

from kubernetes.client import V1ObjectMeta

from .. import CustomResourceDefinition


def to_spark_body(project_name: str, env_vars: dict, spark: dict) -> dict:
    static_body = {
        'type': 'Scala',
        'mode': 'cluster',
        'imagePullPolicy': 'Always',
        'sparkVersion': '3.1.1',
        'restartPolicy': {
            'type': 'Never'
        },
        'sparkConfigMap': project_name,
        'image': 'bigdataregistry.azurecr.io/send-slack-notification:PR-1231',
        'arguments': [
            'python',
            '-m',
            'job_testing_mpl_k8s.main_send_slack_notification'
        ],
        'driver': {
            'cores': 1,
            'coreLimit': '1200m',
            'memory': '5G',
            'memoryOverhead': '1024',
            'labels': {
                'version': '3.1.1'
            },
            'serviceAccount': project_name,
            'envVars': env_vars,
            'envSecretKeyRefs': None
        },
        'executor': {
            'cores': 1,
            'instances': 1,
            'memory': '3G',
            'memoryOverhead': '2048',
            'labels': {
                'version': '3.1.1'
            },
            'envVars': env_vars,
            'envSecretKeyRefs': None
        },
        'deps': {
            'jars': [
                'https://repo1.maven.org/maven2/com/microsoft/sqlserver/'
                'mssql-jdbc/11.2.1.jre8/mssql-jdbc-11.2.1.jre8.jar'
            ]
        },
        'sparkConf': {
            'spark.driver.extraClassPath': 'mssql-jdbc-11.2.1.jre8.jar',
            'spark.executor.extraClassPath': 'mssql-jdbc-11.2.1.jre8.jar',
            'spark.sql.legacy.timeParserPolicy': 'LEGACY',
            'spark.sql.broadcastTimeout': '600'
        }
    }

    return static_body | spark


def get_spark_config_map_data() -> dict:
    return {
        'log4j.properties':
            '#\n'
            '# Licensed to the Apache Software Foundation (ASF) under one or more\n'
            '# contributor license agreements.  See the NOTICE file distributed with\n'
            '# this work for additional information regarding copyright ownership.\n'
            '# The ASF licenses this file to You under the Apache License, Version 2.0\n'
            '# (the "License"); you may not use this file except in compliance with\n'
            '# the License.  You may obtain a copy of the License at\n'
            '#\n'
            '#    http://www.apache.org/licenses/LICENSE-2.0\n'
            '#\n'
            '# Unless required by applicable law or agreed to in writing, software\n'
            '# distributed under the License is distributed on an "AS IS" BASIS,\n'
            '# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.\n'
            '# See the License for the specific language governing permissions and\n'
            '# limitations under the License.\n'
            '#\n'
            '#log everything to file at the beginning\n'
            'log4j.rootCategory=INFO, console, sparklog\n'
            '#afterwards log only our messages to the console\n'
            'log4j.logger.nl.vandebron.sparkplug=INFO, console\n'
            '#configure ConsoleAppender\n'
            'log4j.appender.console=org.apache.log4j.ConsoleAppender\n'
            'log4j.appender.console.target=System.out\n'
            'log4j.appender.console.layout=org.apache.log4j.PatternLayout\n'
            'log4j.appender.console.layout.ConversionPattern=%d{yy/MM/dd HH:mm:ss} %p %c{1}: %m%n\n'
            'log4j.appender.sparklog=org.apache.log4j.FileAppender\n'
            '#this must be overwritten programmatically when setting the appender\n'
            'log4j.appender.sparklog.File=/dev/null\n'
            'log4j.appender.sparklog.layout=org.apache.log4j.PatternLayout\n'
            'log4j.appender.sparklog.layout.ConversionPattern=%d %-5p [%c{1}] %m%n\n'
            'log4j.appender.sparklog.Threshold=INFO\n'
            'log4j.appender.sparklog.Append=true\n'
            '# Set the default spark-shell log level to WARN. When running the spark-shell, the\n'
            '# log level for this class is used to overwrite the root logger\'s log level, so that\n'
            '# the user can have different defaults for the shell and regular Spark apps.\n'
            'log4j.logger.org.apache.spark.repl.Main=WARN\n'
            '# Settings to quiet third party logs that are too verbose\n'
            'log4j.logger.org.spark_project.jetty=WARN\n'
            'log4j.logger.org.spark_project.jetty.util.component.AbstractLifeCycle=ERROR\n'
            'log4j.logger.org.apache.spark.repl.SparkIMain$exprTyper=INFO\n'
            'log4j.logger.org.apache.spark.repl.SparkILoop$SparkILoopInterpreter=INFO\n'
            'log4j.logger.org.apache.parquet=ERROR\n'
            'log4j.logger.parquet=ERROR\n'
            '# SPARK-9183: Settings to avoid annoying messages when looking up '
            'nonexistent UDFs in SparkSQL with Hive support\n'
            'log4j.logger.org.apache.hadoop.hive.metastore.RetryingHMSHandler=FATAL\n'
            'log4j.logger.org.apache.hadoop.hive.ql.exec.FunctionRegistry=ERROR',
    }


class V1SparkApplication(CustomResourceDefinition):
    def __init__(self, schedule: Optional[str], body: dict):
        if schedule:
            super().__init__(api_version="sparkoperator.k8s.io/v1beta2", kind="ScheduledSparkApplication",
                             metadata=V1ObjectMeta(name="sparkapplications.sparkoperator.k8s.io"),
                             schema="sparkoperator.k8s.io_scheduledsparkapplications.schema.yml",
                             spec={'schedule': schedule,
                                   'template': body})
        else:
            super().__init__(api_version="sparkoperator.k8s.io/v1beta2", kind="SparkApplication",
                             metadata=V1ObjectMeta(name="sparkapplications.sparkoperator.k8s.io"),
                             schema="sparkoperator.k8s.io_sparkapplications.schema.yml",
                             spec={'spec': body})



"""Utility functions for Jenkins"""
from dataclasses import dataclass

from ...project import Target


@dataclass(frozen=True)
class JenkinsConfig:
    url: str
    pipelines: dict[str, str]
    default: str

    @staticmethod
    def from_config(values: dict):
        jenkins_config = values.get('jenkins')
        if not jenkins_config:
            raise KeyError('jenkins should be defined in config')
        return JenkinsConfig(url=jenkins_config['url'], pipelines=jenkins_config['pipelines'],
                             default=jenkins_config['defaultPipeline'])

    @property
    def default_pipeline(self) -> str:
        return self.pipelines[self.default]


@dataclass(frozen=True)
class Pipeline:
    target: Target
    tag: str
    url: str
    pipeline: str
    body: str
    jenkins_config: JenkinsConfig

    def _to_path(self):
        return f'PR-{self.tag}' if self.target == Target.PULL_REQUEST else self.tag

    def pipeline_location(self) -> str:
        return f'{self.jenkins_config.url}job/{self.jenkins_config.pipelines[self.pipeline].replace(" ", "%20")}/'

    def job_location(self) -> str:
        return f'{self.pipeline_location}job/{self._to_path()}/'

    def job_name(self) -> str:
        return f'{self.jenkins_config.pipelines[self.pipeline]}/{self._to_path()}'

    def build_location(self) -> str:
        return f'{self.pipeline_location()}view/change-requests/job/{self._to_path()}/lastBuild/'

    def human_readable(self) -> str:
        return f'[link={self.url}]#{self.tag}[/link]'



"""Jenkins multibranch pipeline build tool"""
import subprocess
from dataclasses import dataclass

import requests
from github import Github
from jenkinsapi.jenkins import Jenkins
from rich.console import Console
from rich.markdown import Markdown

from ....project import Target
from ....utilities.github import GithubConfig, get_pr_for_branch
from ....utilities.jenkins import JenkinsConfig, Pipeline
from ....utilities.jenkins.runner import JenkinsRunner
from ....utilities.repo import RepoConfig, Repository


@dataclass(frozen=True)
class JenkinsRunParameters:
    jenkins_user: str
    jenkins_password: str
    config: dict
    pipeline: str
    verbose: bool


def get_token(github_config: GithubConfig):
    if github_config.token:
        return github_config.token
    return subprocess.run(['gh', 'auth', 'token'], stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
                          check=True).stdout.decode('utf-8').strip()


def run_jenkins(run_config: JenkinsRunParameters):
    log_console = Console(log_path=False, log_time=False)
    with log_console.status('Fetching Github info.. [blue]>gh pr view[/blue]') as status:
        config = run_config.config
        github_config = GithubConfig(config)
        with Repository(RepoConfig(config)) as git_repo:
            try:
                github = Github(login_or_token=get_token(github_config))

                repo = github.get_repo(github_config.repository)

                if git_repo.main_branch == git_repo.get_branch:
                    status.console.log(f'On main branch ({git_repo.get_branch}), cannot determine which PR to build')
                    return

                pull = get_pr_for_branch(repo, git_repo.get_branch)

                jenkins_config = JenkinsConfig.from_config(config)
                pipeline_info = Pipeline(target=Target.PULL_REQUEST, tag=f'{pull.number}', url=pull.url,
                                         pipeline=run_config.pipeline, body=pull.body, jenkins_config=jenkins_config)

                status.start()
                status.update(f'Fetching Jenkins info for {pipeline_info.human_readable()} ...')

                runner = JenkinsRunner(pipeline=pipeline_info,
                                       jenkins=Jenkins(jenkins_config.url, username=run_config.jenkins_user,
                                                       password=run_config.jenkins_password), status=status)
                runner.run()
            except requests.ConnectionError:
                status.console.bell()
                status.console.log('⚠️ Could not connect. Are you on VPN?')
            except Exception as exc:
                status.console.print(Markdown(f'Unexpected exception: {exc}'))
                if run_config.verbose:
                    status.console.print_exception()
                raise exc



"""
Markdown run result formatters
"""
import operator

from junitparser import TestSuite

from ...project import Stage, Project
from ...steps import Output, ArtifactType
from ...steps.deploy.kubernetes import DEPLOYED_SERVICE_KEY
from ...steps.run import RunResult
from ...steps.steps import StepResult, collect_test_results
from ...utilities.junit import TestRunSummary, sum_suites


def summary_to_markdown(summary: TestRunSummary):
    return f"🧪 {summary.tests} ❌ {summary.failures} " \
           f"💔 {summary.errors} 🙈 {summary.skipped}"


def __add_link_if_service(name: str, output: Output) -> str:
    if output.produced_artifact and output.produced_artifact.artifact_type == ArtifactType.DEPLOYED_HELM_APP:
        url = output.produced_artifact.spec.get(DEPLOYED_SERVICE_KEY)
        if url:
            return f'[{name}]({url}/swagger/index.html)'

    return name


def wrap_project_name(proj: Project, result: list[StepResult]):
    project_name = proj.name
    encapsulation = '_'
    found_result = next((r for r in result if r.project.name == project_name), None)
    if found_result:
        project_name = __add_link_if_service(project_name, found_result.output)
        encapsulation = '*' if found_result.output.success else '~~'

    return f'{encapsulation}{project_name}{encapsulation}'


def __to_oneliner(result: list[StepResult], plan: set[Project]) -> str:
    project_names: list[str] = []
    if plan:
        sorted_plan = sorted(plan, key=operator.attrgetter('name'))
        for proj in sorted_plan:
            project_names.append(wrap_project_name(proj, result))
    else:
        project_names = list(map(lambda r: f'_{r.project.name}_', result))

    return f'{", ".join(project_names)}'


def stage_to_icon(stage: Stage):
    if stage == Stage.BUILD:
        return '🏗️ '
    if stage == Stage.TEST:
        return '🧪'
    if stage == Stage.DEPLOY:
        return '🚀'
    return '➡️'


def markdown_for_stage(run_result: RunResult, stage: Stage):
    step_results: list[StepResult] = run_result.results_for_stage(stage)
    plan: set[Project] = run_result.plan_for_stage(stage)
    if not step_results and not plan:
        return ''

    result = f"{stage_to_icon(stage)}  {__to_oneliner(step_results, plan)}  \n"
    test_results = collect_test_results(step_results)
    if test_results:
        result += to_markdown_test_report(
            test_results) + f' [link]({run_result.run_properties.details.tests_url}) \n'

    return result


def run_result_to_markdown(run_result: RunResult) -> str:
    result: str = f'{run_result.status_line}  \n'
    exception = run_result.exception
    if exception:
        result += f"For _{exception.executor}_ on _{exception.project_name}_ at _{exception.stage}_ \n"
        result += f"\n```\n{exception}\n```\n"

    for stage in Stage:
        result += markdown_for_stage(run_result, stage)

    return result


def to_markdown_test_report(suites: list[TestSuite]):
    total_tests = sum_suites(suites)
    return f"{summary_to_markdown(total_tests)}"



""" Step that deploys ephemeral the docker images produced in the build stage """

from logging import Logger

from python_on_whales import docker

from .. import Step, Meta
from ..models import Input, Output, ArtifactType
from ...project import Stage
from ...utilities.docker import docker_image_tag
from ...utilities.ephemeral import get_env_variables


class EphemeralDockerDeploy(Step):

    def __init__(self, logger: Logger) -> None:
        super().__init__(logger, Meta(
            name='Deploy From Docker Container',
            description='Runs and removes the docker container built during the build stage. '
                        'Useful for custom deploy steps depending on technology not bundled with MPyL',
            version='0.0.1',
            stage=Stage.DEPLOY
        ), produced_artifact=ArtifactType.NONE, required_artifact=ArtifactType.DOCKER_IMAGE)

    def execute(self, step_input: Input) -> Output:
        env_variables = get_env_variables(step_input.project, step_input.run_properties.target)
        docker_image_name = docker_image_tag(step_input)
        docker_run_result = docker.run(image=docker_image_name, remove=True, envs=env_variables)
        self._logger.info(docker_run_result)

        return Output(success=True, message=f"Deployed project {step_input.project.name}", produced_artifact=None)



"""Utilities related to launching a subprocess"""

import subprocess
from logging import Logger
from typing import Union

from ...steps.models import Output


def custom_check_output(logger: Logger, command: Union[str, list[str]], capture_stdout: bool = False) -> Output:
    """
    Wrapper around subprocess.Popen
    ⚠️ Using this function implies an implicit runtime OS dependency.
    Avoid this if at all possible. For example, to run docker commands, use the bundled docker client (python-on-whales)
    which makes the dependency on docker explicit and adds a lot of convenience methods.
    """
    if isinstance(command, str):
        command = command.split(' ')

    command_argument = ' '.join(command)
    logger.info(f"Executing: '{command_argument}'")
    try:
        if capture_stdout:
            out = subprocess.check_output(command, stderr=subprocess.STDOUT).decode("utf-8")
            print(out)
            return Output(success=True, message=out)

        with subprocess.Popen(command, stdout=subprocess.PIPE, text=True) as process:
            if not process.stdout:
                raise RuntimeError(f'Process {command_argument} does not have an stdout')

            for line in iter(process.stdout.readline, ""):
                if line:
                    print(line.rstrip())
                if process.poll() is not None:
                    break
            success = process.wait() == 0
            if not success:
                logger.warning(f"Subprocess failed: {process.stderr.read() if process.stderr else 'No stderr output'}")

            return Output(success=success, message='Subprocess executed successfully')


    except subprocess.CalledProcessError as exc:
        logger.warning(f"'{command_argument}': failed with return code: {exc.returncode} err: "
                       f"{exc.stderr.decode() if exc.stderr else 'No stderr output'}",
                       exc_info=True)

    except FileNotFoundError:
        logger.warning(f"'{command_argument}: file not found", exc_info=True)

    return Output(success=False, message='Subprocess failed')



"""Simple MPyL build runner"""

import logging
from dataclasses import dataclass
from pathlib import Path
from typing import Optional

from jsonschema import ValidationError
from rich.console import Console
from rich.logging import RichHandler
from rich.markdown import Markdown

from ....project import load_project, Stage, Project
from ....reporting.formatting.markdown import run_result_to_markdown
from ....reporting.targets import Reporter
from ....stages.discovery import for_stage, find_invalidated_projects_per_stage
from ....steps.models import RunProperties
from ....steps.run import RunResult
from ....steps.steps import Steps, ExecutionException
from ....utilities.repo import Repository, RepoConfig


@dataclass(frozen=True)
class MpylRunConfig:
    config: dict
    run_properties: RunProperties


@dataclass(frozen=True)
class MpylCliParameters:
    local: bool
    pull_main: bool = False
    verbose: bool = False
    all: bool = False


@dataclass(frozen=True)
class MpylRunParameters:
    run_config: MpylRunConfig
    parameters: MpylCliParameters


FORMAT = "%(name)s  %(message)s"


def get_build_plan(logger: logging.Logger, repo: Repository, mpyl_run_parameters: MpylRunParameters) -> RunResult:
    params = mpyl_run_parameters.parameters
    logger.info(f"Running with {params}")
    if params.pull_main:
        if repo.main_branch_pulled:
            logger.info(f'Branch {repo.main_branch} already present locally. Skipping pull.')
        else:
            logger.info(f'Pulling {repo.main_branch} from {repo.get_remote_url}')
            pull_result = repo.pull_main_branch()
            logger.info(f'Pulled `{pull_result[0].remote_ref_path.strip()}` to local')

    changes_in_branch = repo.changes_in_branch_including_local() if params.local else repo.changes_in_branch()
    logger.debug(f'Changes: {changes_in_branch}')

    projects_per_stage: dict[Stage, set[Project]] = find_build_set(repo, changes_in_branch, params.all)
    return RunResult(run_properties=mpyl_run_parameters.run_config.run_properties, run_plan=projects_per_stage)


def run_mpyl(mpyl_run_parameters: MpylRunParameters, reporter: Optional[Reporter]) -> RunResult:
    params = mpyl_run_parameters.parameters
    console = Console(markup=True, width=None if params.local else 135, no_color=False, log_path=False,
                      color_system='256')
    logging.basicConfig(
        level="DEBUG" if params.verbose else "INFO", format=FORMAT, datefmt="[%X]",
        handlers=[RichHandler(markup=True,
                              console=console, show_path=params.local)]
    )
    logger = logging.getLogger('mpyl')
    try:
        with Repository(RepoConfig(mpyl_run_parameters.run_config.config)) as repo:

            run_plan = get_build_plan(logger, repo, mpyl_run_parameters)

            if not run_plan.run_plan.items():
                logger.info("Nothing to do. Exiting..")
                return run_plan

            logger.info("Building plan:")
            console.print(Markdown(f"\n\n{run_result_to_markdown(run_plan)}"))

            run_result: RunResult = run_plan
            try:
                steps = Steps(logger=logger, properties=mpyl_run_parameters.run_config.run_properties)
                run_result = run_build(run_plan, steps, reporter, mpyl_run_parameters.parameters.local)
            except ValidationError as exc:
                console.log(f'Schema validation failed {exc.message} at `{".".join(map(str, exc.path))}`')
                raise exc
            except ExecutionException as exc:  # pylint: disable=broad-except
                run_result.exception = exc
                console.log(f'Exception during build execution: {exc}')
                console.print_exception()

            console.print(Markdown(run_result_to_markdown(run_result)))
            return run_result

    except Exception as exc:
        console.log(f'Unexpected exception: {exc}')
        console.print_exception()
        raise exc


def find_build_set(repo: Repository, changes_in_branch, build_all: bool) -> dict[Stage, set[Project]]:
    project_paths = repo.find_projects()
    all_projects = set(map(lambda p: load_project(Path(""), Path(p), False), project_paths))

    if build_all:
        return {Stage.BUILD: for_stage(all_projects, Stage.BUILD),
                Stage.TEST: for_stage(all_projects, Stage.TEST),
                Stage.DEPLOY: for_stage(all_projects, Stage.DEPLOY)}

    return find_invalidated_projects_per_stage(all_projects, changes_in_branch)


def run_build(accumulator: RunResult, executor: Steps, reporter: Optional[Reporter] = None, dry_run: bool = True):
    for stage, projects in accumulator.run_plan.items():
        for proj in projects:
            result = executor.execute(stage, proj, dry_run)
            accumulator.append(result)
            if reporter:
                reporter.send_report(accumulator)

            if not result.output.success:
                logging.warning(f'Build failed at {stage} for {proj.name}')
                return accumulator
    return accumulator



"""
Custom Resource Definitions, allow you to create custom k8s resources, besides the default ones.
This is useful for example when you want to configure a specific operator, like Spark or sealed secrets.
"""
import pkgutil
from io import StringIO
from typing import Optional

import jsonschema
import six
from jsonschema import ValidationError
from kubernetes.client import Configuration, V1ObjectMeta
from ruamel.yaml import YAML

yaml = YAML()


class CustomResourceDefinition:
    openapi_types = {
        'api_version': 'str',
        'kind': 'str',
        'metadata': 'V1ObjectMeta',
        'spec': 'dict'
    }

    attribute_map = {
        'api_version': 'apiVersion',
        'kind': 'kind',
        'metadata': 'metadata',
        'spec': 'spec'
    }

    def __init__(self, api_version: str, kind: str, metadata: V1ObjectMeta, spec: dict,
                 local_vars_configuration=None, schema: Optional[str] = None):  # noqa: E501
        """V1CSIDriver - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._api_version = api_version
        self._kind = kind
        self._metadata = metadata
        self._spec = spec
        self._schema = schema
        self.discriminator = None

        if api_version is not None:
            self.api_version = api_version
        if kind is not None:
            self.kind = kind
        if metadata is not None:
            self.metadata = metadata

    @property
    def schema(self) -> Optional[str]:
        return self._schema

    @property
    def api_version(self) -> str:
        return self._api_version

    @api_version.setter
    def api_version(self, api_version):
        self._api_version = api_version

    @property
    def kind(self) -> str:
        return self._kind

    @kind.setter
    def kind(self, kind):
        self._kind = kind

    @property
    def metadata(self) -> V1ObjectMeta:
        return self._metadata

    @metadata.setter
    def metadata(self, metadata):
        self._metadata = metadata

    @property
    def spec(self) -> dict:
        return self._spec

    @spec.setter
    def spec(self, spec):
        if self.local_vars_configuration.client_side_validation and spec is None:  # noqa: E501
            raise ValueError("Invalid value for `spec`, must not be `None`")  # noqa: E501
        self._spec = spec


def to_dict(obj):
    result = {}

    for attr, _ in six.iteritems(obj.openapi_types):
        value = getattr(obj, attr)
        key = obj.attribute_map.get(attr)
        if isinstance(value, list):
            result[key] = list(map(
                lambda x: to_dict(x) if hasattr(x, "to_dict") else x,
                value
            ))
        elif hasattr(value, "to_dict"):
            result[key] = to_dict(value)
        elif isinstance(value, dict):
            result[key] = dict(map(
                lambda item: (item[0], to_dict(item[1]))
                if hasattr(item[1], "to_dict") else item,
                value.items()
            ))
        else:
            result[key] = value

    return result


def to_yaml(resource: object) -> str:
    def remove_none(obj):
        if isinstance(obj, (list, tuple, set)):
            return type(obj)(remove_none(x) for x in obj if x is not None)
        if isinstance(obj, dict):
            return type(obj)((remove_none(k), remove_none(v))
                             for k, v in obj.items() if k is not None and v is not None)
        return obj

    resource_dict = to_dict(resource) if (
            hasattr(resource, "openapi_types") and hasattr(resource, "attribute_map")) else {}
    yaml_values = remove_none(resource_dict)

    if hasattr(resource, 'schema') and resource.schema:
        template = pkgutil.get_data(__name__, f'schema/{resource.schema}')
        if template:
            schema = yaml.load(template.decode('utf-8'))
            try:
                jsonschema.validate(yaml_values, schema)
            except ValidationError as err:
                raise ValueError(
                    f'Schema validation failed with {err.message} at {".".join(map(str, err.schema_path))}') from err
        else:
            raise ValueError(f'Schema {resource.schema} defined but not found in package')

    stream = StringIO()
    yaml.dump(yaml_values, stream)

    return stream.getvalue()



"""
## What is a Step?
A `Step` is a unit of execution within a pipeline run. Each step relates to one particular `mpyl.project.Stage`.
For example, there are some built in steps in the `Build` stage like `mpyl.steps.build.dockerbuild.BuildDocker` or
`mpyl.steps.build.sbt.BuildSbt`. Both of these steps compile and assemble source code into a bundle that is embedded
in a docker image.


## How do I create my own custom step?

### Before you start
A step can either be *built in* to `MPyL` or defined within the context of a particular pipeline. The latter is easier
to do and is the best option if the likelihood of the logic being of use in other contexts is small.

.. note:: The benefits of standard artifact types
   It is recommended to have your steps produce any of the standard artifact types. This allows you to benefit from
   the tooling embedded in MPyL to process them. For example: if a step in the `Test` stage produces an artifact of type
   `ArtifactType.JUNIT_TESTS`, the test results can be reported in a unified way by any of the `mpyl.reporting`
   reporters regardless of whether the tests were run by a `jest` or a `scala` test.


### Implementation

Let's assume you want to create a new step for the `Build` stage that creates a builds a `Java` maven project and embeds
it a docker image.

Implement the `Step` interface as `BuildJava(Step)`, or however you want to call it. For an example you could have
 a look at the `mpyl.steps.build.echo.BuildEcho`.

##### Important constructor properties
 - `Meta.name` is how you can refer to this step from a `project.yml`. The `mpyl.steps.steps.Steps` executor will pick
 up the `Step` implementation that has the name described in `stages.build` to execute the build step.
```yaml
name: 'javaService'
stages:
    build: 'Java Build'
description: 'A simple Java service'
```
 -  `Meta.stage` describes the `mpyl.project.Stage to which the step relates. It can only be executed in this context.
 -  `Step.produced_artifact` defines the `mpyl.steps.models.ArtifactType` that this step produces. In our example case
 this would be `mpyl.steps.models.ArtifactType.DOCKER_IMAGE`.
 - The `Step.after` is a postprocessing step, which we can set to `mpyl.steps.build.docker_after_build.AfterBuildDocker`
 in this case. It will push the image produced by this step to a registry.

##### Return type

Your step needs to return an `mpyl.steps.models.Output` object with fields that are hopefully self-explanatory.
The `produced_artifact` can be constructed with `mpyl.steps.models.input_to_artifact`. The `spec` parameter is for
specifying `ArtifactType` specific metadata, that can be picked up by another step that has `ArtifactType.DOCKER_IMAGE`
as `required_artifact`, like `mpyl.steps.deploy.kubernetes.DeployKubernetes` for example.
```python
input_to_artifact(ArtifactType.DOCKER_IMAGE, step_input, spec={'image': image_tag})
```

##### Step input
The step receives an `mpyl.steps.models.Input` for `execute`. If your step needs configuration settings, like for
example `mpyl.utilities.docker.DockerConfig`, this can be constructed from the `mpyl.steps.models.RunProperties.config`
dictionary on `mpyl.steps.models.Input.run_properties`.
Make sure to update the schema under `src/mpyl/schema/mpyl_config.schema.yml` accordingly, so that the configuration
remains type safe and mistakes are found as early as possible.
"""
from __future__ import annotations

from dataclasses import dataclass
from logging import Logger
from typing import Optional, List

from .models import ArtifactType, Input, Output
from ..project import Stage


class IPluginRegistry(type):
    plugin_registries: List[type] = []

    def __init__(cls, name, _bases, _attrs):
        super().__init__(cls)
        if name != 'Step':
            IPluginRegistry.plugin_registries.append(cls)


@dataclass(frozen=True)
class Meta:
    name: str
    """External, unique identifier. The step can be referred to by this name from `project.yml`"""
    description: str
    version: str
    stage: Stage
    """The stage that this step relates to"""

    def __str__(self) -> str:
        return f'{self.name}: {self.version}'


class Step(metaclass=IPluginRegistry):
    """ Abstract base class for execution steps. Any execution step (e.g. build, test, deploy) will need to implement
    this interface.
    """
    meta: Meta
    """Information _about_ the specific instance of `Step`. For example its name, description, version or the stage
    to which it applies.
    """
    produced_artifact: ArtifactType
    """The type of the artifact produced by this step """
    required_artifact: ArtifactType
    """Is set to something other than `ArtifactType.NONE` if this step depends on an artifact produced by the execution
    of an earlier step. For example: a step in the `Deploy` stage, may need to deploy a docker image that was produced
    in the `Build` stage."""
    before: Optional[Step]
    after: Optional[Step]
    """Will be executed after completion of this step. Can be used for shared post processing steps, like pushing the
    produced docker image to a registry or filing test results."""

    def __init__(self, logger: Logger, meta: Meta, produced_artifact: ArtifactType,
                 required_artifact: ArtifactType, before: Optional[Step] = None, after: Optional[Step] = None) -> None:
        self._logger = logger.getChild(meta.name.replace(' ', ''))
        self.meta = meta
        self.produced_artifact = produced_artifact
        self.required_artifact = required_artifact
        self.before = before
        self.after = after

    def execute(self, step_input: Input) -> Output:
        """ Execute an individual step for a specific `project` at a specific `stage` of the pipeline.
        :param step_input: The input of the project along with its build properties and required artifact (if any).
        :return Output: The result of the execution. `success` will be `False` if any exception was thrown during
        execution.
        """
        return Output(success=False, message=f"Not implemented for {step_input.project.name}", produced_artifact=None)



"""Projects and how they relate to each other"""
from dataclasses import dataclass
from enum import Enum
from typing import Optional

from ..project import Project


@dataclass(frozen=True)
class Protocol(Enum):
    UNKNOWN = 1
    REST = 2
    KAFKA = 3


@dataclass(frozen=True)
class Contract:
    type: Protocol
    contract_file: Optional[str]


@dataclass(frozen=True)
class Dependency:
    project: Project
    contracts: set[Contract]


@dataclass
class ProjectWithDependents:
    project: Project
    dependent_projects: dict[str, Dependency]

    @property
    def name(self):
        return self.project.name



""" Dummy test step to test the framework. """

from logging import Logger
from pathlib import Path

from .. import Step, Meta
from ..models import Input, Output, ArtifactType, input_to_artifact
from ...project import Stage
from ...utilities.junit import TEST_OUTPUT_PATH_KEY

SAMPLE_JUNIT_RESULT = """
<?xml version="1.0" encoding="UTF-8"?>
<testsuites name="jest tests" tests="1" failures="0" errors="0" time="0.486">
  <testsuite name="undefined" errors="0" failures="0" skipped="0" timestamp="2023-03-29T02:51:00" time="0.244" tests="1">
    <testcase classname="Test Echo" name=" fake test case from TestEcho step" time="0.002">
    </testcase>
  </testsuite>
</testsuites>
""".strip()


class TestEcho(Step):

    def __init__(self, logger: Logger) -> None:
        super().__init__(logger, Meta(
            name='Echo Test',
            description='Dummy test step to test the framework',
            version='0.0.1',
            stage=Stage.TEST
        ), produced_artifact=ArtifactType.JUNIT_TESTS, required_artifact=ArtifactType.NONE)

    def execute(self, step_input: Input) -> Output:
        self._logger.info(f"Testing project {step_input.project.name}")
        path = Path(step_input.project.target_path, "test_results")
        path.mkdir(parents=True, exist_ok=True)
        Path(path, "test.xml").write_text(SAMPLE_JUNIT_RESULT, encoding='utf-8')

        artifact = input_to_artifact(artifact_type=ArtifactType.JUNIT_TESTS, step_input=step_input,
                                     spec={TEST_OUTPUT_PATH_KEY: str(path)})
        return Output(success=True, message=f"Tested {step_input.project.name}", produced_artifact=artifact)



"""
Accumulate `mpyl.steps.run.RunResult` from executed `mpyl.steps.step.Step`
"""

import operator
from typing import Optional

from .models import RunProperties
from ..project import Stage, Project
from .steps import StepResult, ExecutionException


class RunResult:
    _run_plan: dict[Stage, set[Project]]
    _results: list[StepResult]
    _run_properties: RunProperties
    _exception: Optional[ExecutionException]

    def __init__(self, run_properties: RunProperties, run_plan=None):
        if run_plan is None:
            run_plan = {}
        self._run_properties = run_properties
        self._run_plan = run_plan
        self._exception = None
        self._results = []

    @property
    def status_line(self) -> str:
        if self._exception:
            return '❗ Failed with exception'
        if self.is_in_progress:
            return '🏗️ Building'
        if not self.has_results:
            return '🦥 Nothing to do'
        if self._results_success():
            return '✅ Successful'

        return '❌ Failed'

    @property
    def progress_fraction(self) -> float:
        unfinished = 0
        finished = 0
        for stage, projects in self.run_plan.items():
            finished_project_names = set(map(lambda r: r.project.name, self.results_for_stage(stage)))
            for project in projects:
                if project.name in finished_project_names:
                    finished += 1
                else:
                    unfinished += 1

        total = unfinished + finished
        if total == 0:
            return 0.0

        return 1.0 - (unfinished / total)

    @property
    def exception(self) -> Optional[ExecutionException]:
        return self._exception

    @exception.setter
    def exception(self, exception: ExecutionException):
        self._exception = exception

    @property
    def run_properties(self) -> RunProperties:
        return self._run_properties

    @property
    def run_plan(self) -> dict[Stage, set[Project]]:
        return self._run_plan

    def append(self, result: StepResult):
        self._results.append(result)

    def extend(self, results: list[StepResult]):
        self._results.extend(results)

    @property
    def is_success(self):
        if self._exception:
            return False
        return self._results_success()

    @property
    def is_finished(self):
        return self.progress_fraction == 1.0

    @property
    def has_results(self):
        return len(self._results) > 0

    @property
    def is_in_progress(self):
        return self.is_success and not self.is_finished

    def _results_success(self):
        return not self.has_results or all(r.output.success for r in self._results)

    @staticmethod
    def sort_chronologically(results: list[StepResult]) -> list[StepResult]:
        return sorted(results, key=operator.attrgetter('timestamp'))

    def results_for_stage(self, stage: Stage) -> list[StepResult]:
        return RunResult.sort_chronologically([res for res in self._results if res.stage == stage])

    def plan_for_stage(self, stage: Stage) -> set[Project]:
        plan: Optional[set[Project]] = self.run_plan.get(stage)
        if plan:
            return plan

        return set()



""" Loads all projects inside a repository. """
from pathlib import Path
from typing import Optional

from . import ProjectWithDependents, Protocol, Contract, Dependency
from ..project import Project, load_project, Stage


def load_projects(root_dir: Path, paths: list[str], strict: bool = False) -> set[Project]:
    return set(map(lambda p: load_project(root_dir, Path(p), strict), paths))


def find_by_contract_dep(dep: str, projects: list[ProjectWithDependents]) -> Optional[ProjectWithDependents]:
    for project in projects:
        if project.project.root_path.replace('./', '') in dep:
            return project
    return None


def find_dependencies(project: Project, other_projects: set[Project]) -> ProjectWithDependents:
    mapped = list(map(lambda p: ProjectWithDependents(p, {}), other_projects))

    def recursively_find(proj: ProjectWithDependents, other: list[ProjectWithDependents], discovery_stack: list[str]):
        dependent_projects = {}
        for dep in proj.project.dependencies.set_for_stage(Stage.TEST) if proj.project.dependencies else set():
            found = find_by_contract_dep(dep, other)
            if found:
                if found.name not in discovery_stack:
                    discovery_stack.append(found.name)
                    recursively_find(found, other, discovery_stack[:])
                    dependent_projects[found.name] = Dependency(found.project, {Contract(Protocol.UNKNOWN, dep)})
                elif found.name in dependent_projects:
                    dependent_projects[found.name].contracts.add(Contract(Protocol.UNKNOWN, dep))

        proj.dependent_projects = dependent_projects
        return proj

    return recursively_find(ProjectWithDependents(project=project, dependent_projects={}), mapped, [])


def find_all_dependencies(root_dir: Path, paths: list[str]) -> list[ProjectWithDependents]:
    projects = load_projects(root_dir, paths)
    return list(map(lambda p: find_dependencies(p, projects), projects))



"""
.. include:: ../../README.md
<details>
  <summary>Mission statement</summary>
.. include:: ../../README-motivation.md
</details>

## Technologies

### Requirements
The following technologies are expected to be present on the local OS:
 - [Python](https://www.python.org/) >= 3.9
 - [Pip](https://pypi.org/project/pip/) >= 23.0.1
 - [Docker](https://www.docker.com/) > 20
 - [Docker compose](https://docs.docker.com/compose/install/linux/)
 installed as plugin (`docker compose version`) >= v2.2.3
 - [Git](https://git-scm.com/) SCM

### Bundled
MPyL is extensible and has a minimal footprint. Having said that, batteries for the following technologies are included.

##### CI/CD
###### Build
 - [Docker](https://www.docker.com/) `mpyl.steps.build.dockerbuild`
 - [Scala (SBT)](https://www.scala-sbt.org/) `mpyl.steps.build.sbt`
 - [Jenkins](https://www.jenkins.io/)  `mpyl.cli.commands.build.jenkins` and
 `mpyl.utilities.jenkins.runner.JenkinsRunner`

###### Testing
 - [Junit](https://junit.org/) `mpyl.steps.models.ArtifactType.JUNIT_TESTS`

###### Deployment
 - [K8S](https://kubernetes.io/) `mpyl.steps.deploy.kubernetes`
 - [Helm](https://helm.sh/) `mpyl.steps.deploy.k8s.helm`

##### Reporting
 - [Jira](https://www.atlassian.com) `mpyl.reporting.targets.jira`
 - [Github](https://github.com/) `mpyl.reporting.targets.github`
 - [Slack](https://slack.com/) `mpyl.reporting.targets.slack`

.. include:: ../../README-usage.md

.. include:: ../../README-dev.md
"""

import click

from .cli.build import build
from .cli.health import health
from .cli.meta_info import get_version
from .cli.meta_info import version
from .cli.projects import projects
from .utilities.pyaml_env import parse_config
from .utilities.repo import RepoConfig, Repository


@click.group(name='mpyl')
def main_group():
    """Command Line Interface for MPyL"""


def main():
    main_group.help = f"Command Line Interface for MPyL {get_version()}"
    main_group.add_command(projects)
    main_group.add_command(build)
    main_group.add_command(version)
    main_group.add_command(health)
    main_group()  # pylint: disable = no-value-for-parameter



""" Step that deploys the docker image produced in the build stage to Kubernetes, using HELM. """
import re
from logging import Logger
from typing import Optional

from .k8s import deploy_helm_chart, CustomResourceDefinition
from .k8s.chart import ChartBuilder, to_service_chart
from .. import Step, Meta
from ..models import Input, Output, ArtifactType, input_to_artifact
from ...project import Stage

DEPLOYED_SERVICE_KEY = 'url'


class DeployKubernetes(Step):

    def __init__(self, logger: Logger) -> None:
        super().__init__(logger, Meta(
            name='Kubernetes Deploy',
            description='Deploy to k8s',
            version='0.0.1',
            stage=Stage.DEPLOY
        ), produced_artifact=ArtifactType.NONE, required_artifact=ArtifactType.DOCKER_IMAGE)

    @staticmethod
    def try_extract_endpoint(chart: dict[str, CustomResourceDefinition]) -> Optional[str]:
        ingress = chart.get('ingress-https-route')
        if ingress:
            routes = ingress.spec.get('routes', {})
            if routes:
                url = routes[0].get('match')
                return 'https://' + next(iter(re.findall(r'`(.*)`', url)))
        return None

    def execute(self, step_input: Input) -> Output:
        builder = ChartBuilder(step_input)
        chart = to_service_chart(builder)

        deploy_result = deploy_helm_chart(self._logger, chart, step_input, builder.release_name)
        if deploy_result.success:
            endpoint = self.try_extract_endpoint(chart)
            spec = {}
            if endpoint:
                self._logger.info(f"Service {step_input.project.name} reachable at: {endpoint}")
                spec[DEPLOYED_SERVICE_KEY] = endpoint
            artifact = input_to_artifact(ArtifactType.DEPLOYED_HELM_APP, step_input, spec=spec)
            return Output(success=True, message=deploy_result.message, produced_artifact=artifact)

        return deploy_result



"""Health check command"""

import click
from rich.console import Console

from . import create_console_logger
from .commands.health.checks import perform_health_checks


@click.command('health')
def health():
    """Health check"""
    console: Console = create_console_logger(local=False, verbose=False)
    perform_health_checks(console)



""" Dummy build step to test the framework. """

from logging import Logger

from .. import Step, Meta
from ..models import Input, Output, ArtifactType
from ...project import Stage


class BuildEcho(Step):

    def __init__(self, logger: Logger) -> None:
        super().__init__(logger, Meta(
            name='Echo Build',
            description='Dummy build step to test the framework',
            version='0.0.1',
            stage=Stage.BUILD
        ), produced_artifact=ArtifactType.NONE, required_artifact=ArtifactType.NONE)

    def execute(self, step_input: Input) -> Output:
        self._logger.info(f"Building project {step_input.project.name}")
        return Output(success=True, message=f"Built {step_input.project.name}", produced_artifact=None)
